# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

# Database Module
# --------------------

from __future__ import unicode_literals

def setup_database(force, source_sql=None, verbose=None):
	import frappe
	if frappe.conf.db_type == 'postgres':
		import frappe.database.postgres.setup_db
		return frappe.database.postgres.setup_db.setup_database(force, source_sql, verbose)
	else:
		import frappe.database.mariadb.setup_db
		return frappe.database.mariadb.setup_db.setup_database(force, source_sql, verbose)

def drop_user_and_database(db_name, root_login=None, root_password=None):
	import frappe
	if frappe.conf.db_type == 'postgres':
		pass
	else:
		import frappe.database.mariadb.setup_db
		return frappe.database.mariadb.setup_db.drop_user_and_database(db_name, root_login, root_password)

def get_db(host=None, user=None, password=None, port=None):
	import frappe
	if frappe.conf.db_type == 'postgres':
		import frappe.database.postgres.database
		return frappe.database.postgres.database.PostgresDatabase(host, user, password, port=port)
	else:
		import frappe.database.mariadb.database
		return frappe.database.mariadb.database.MariaDBDatabase(host, user, password, port=port)

def setup_help_database(help_db_name):
	import frappe
	if frappe.conf.db_type == 'postgres':
		import frappe.database.postgres.setup_db
		return frappe.database.postgres.setup_db.setup_help_database(help_db_name)
	else:
		import frappe.database.mariadb.setup_db
		return frappe.database.mariadb.setup_db.setup_help_database(help_db_name)



import os
import frappe


class DbManager:

	def __init__(self, db):
		"""
		Pass root_conn here for access to all databases.
		"""
		if db:
			self.db = db

	def get_current_host(self):
		return self.db.sql("select user()")[0][0].split('@')[1]

	def create_user(self, user, password, host=None):
		# Create user if it doesn't exist.
		if not host:
			host = self.get_current_host()

		if password:
			self.db.sql("CREATE USER '%s'@'%s' IDENTIFIED BY '%s';" % (user, host, password))
		else:
			self.db.sql("CREATE USER '%s'@'%s';" % (user, host))

	def delete_user(self, target, host=None):
		if not host:
			host = self.get_current_host()
		try:
			self.db.sql("DROP USER '%s'@'%s';" % (target, host))
		except Exception as e:
			if e.args[0] == 1396:
				pass
			else:
				raise

	def create_database(self, target):
		if target in self.get_database_list():
			self.drop_database(target)

		self.db.sql("CREATE DATABASE `%s` ;" % target)

	def drop_database(self, target):
		self.db.sql("DROP DATABASE IF EXISTS `%s`;" % target)

	def grant_all_privileges(self, target, user, host=None):
		if not host:
			host = self.get_current_host()

		if frappe.conf.get('rds_db', 0) == 1:
			self.db.sql("GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP, INDEX, ALTER, CREATE TEMPORARY TABLES, CREATE VIEW, EVENT, TRIGGER, SHOW VIEW, CREATE ROUTINE, ALTER ROUTINE, EXECUTE ON `%s`.* TO '%s'@'%s';" % (target, user, host))
		else:
			self.db.sql("GRANT ALL PRIVILEGES ON `%s`.* TO '%s'@'%s';" % (target, user, host))

	def flush_privileges(self):
		self.db.sql("FLUSH PRIVILEGES")

	def get_database_list(self):
		"""get list of databases"""
		return [d[0] for d in self.db.sql("SHOW DATABASES")]

	@staticmethod
	def restore_database(target, source, user, password):
		from frappe.utils import make_esc
		esc = make_esc('$ ')

		from distutils.spawn import find_executable
		pipe = find_executable('pv')
		if pipe:
			pipe = '{pipe} {source} |'.format(
				pipe=pipe,
				source=source
			)
			source = ''
		else:
			pipe = ''
			source = '< {source}'.format(source=source)

		if pipe:
			print('Creating Database...')

		command = '{pipe} mysql -u {user} -p{password} -h{host} {target} {source}'.format(
			pipe=pipe,
			user=esc(user),
			password=esc(password),
			host=esc(frappe.db.host),
			target=esc(target),
			source=source
		)
		os.system(command)



import re, datetime, math, time
from dateutil import parser
import frappe
from six import iteritems, text_type, string_types, integer_types
import babel.dates
from babel.core import UnknownLocaleError
import operator, os

DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S.%f"
DATETIME_FORMAT = DATE_FORMAT + " " + TIME_FORMAT


def flt(s, precision=None):
    """Convert to float (ignore commas)"""
    if isinstance(s, string_types):
        s = s.replace(',', '')

    try:
        num = float(s)
        if precision is not None:
            num = rounded(num, precision)
    except Exception:
        num = 0

    return num


def rounded(num, precision=0):
    """round method for round halfs to nearest even algorithm aka banker's rounding - compatible with python3"""
    precision = cint(precision)
    multiplier = 10 ** precision

    # avoid rounding errors
    num = round(num * multiplier if precision else num, 8)

    floor = math.floor(num)
    decimal_part = num - floor

    if not precision and decimal_part == 0.5:
        num = floor if (floor % 2 == 0) else floor + 1
    else:
        if decimal_part == 0.5:
            num = floor + 1
        else:
            num = round(num)

    return (num / multiplier) if precision else num


def cint(s):
    """Convert to integer"""
    try:
        num = int(float(s))
    except:
        num = 0
    return num


def cast_fieldtype(fieldtype, value):
    if fieldtype in ("Currency", "Float", "Percent"):
        value = flt(value)

    elif fieldtype in ("Int", "Check"):
        value = cint(value)

    elif fieldtype in ("Data", "Text", "Small Text", "Long Text",
                       "Text Editor", "Select", "Link", "Dynamic Link"):
        value = cstr(value)

    elif fieldtype == "Date":
        value = getdate(value)

    elif fieldtype == "Datetime":
        value = get_datetime(value)

    elif fieldtype == "Time":
        value = to_timedelta(value)

    return value


def cstr(s, encoding='utf-8'):
    return frappe.as_unicode(s, encoding)


def get_filter(doctype, f):
    """Returns a _dict like

        {
            "doctype":
            "fieldname":
            "operator":
            "value":
        }
    """
    from frappe.model import default_fields, optional_fields

    if isinstance(f, dict):
        key, value = next(iter(f.items()))
        f = make_filter_tuple(doctype, key, value)

    if not isinstance(f, (list, tuple)):
        frappe.throw(frappe._("Filter must be a tuple or list (in a list)"))

    if len(f) == 3:
        f = (doctype, f[0], f[1], f[2])
    elif len(f) > 4:
        f = f[0:4]
    elif len(f) != 4:
        frappe.throw(frappe._("Filter must have 4 values (doctype, fieldname, operator, value): {0}").format(str(f)))

    f = frappe._dict(doctype=f[0], fieldname=f[1], operator=f[2], value=f[3])

    sanitize_column(f.fieldname)

    if not f.operator:
        # if operator is missing
        f.operator = "="

    valid_operators = ("=", "!=", ">", "<", ">=", "<=", "like", "not like", "in", "not in", "is",
                       "between", "descendants of", "ancestors of", "not descendants of", "not ancestors of",
                       "previous", "next")
    if f.operator.lower() not in valid_operators:
        frappe.throw(frappe._("Operator must be one of {0}").format(", ".join(valid_operators)))

    if f.doctype and (f.fieldname not in default_fields + optional_fields):
        # verify fieldname belongs to the doctype
        meta = frappe.get_meta(f.doctype)
        if not meta.has_field(f.fieldname):

            # try and match the doctype name from child tables
            for df in meta.get_table_fields():
                if frappe.get_meta(df.options).has_field(f.fieldname):
                    f.doctype = df.options
                    break

    return f


def lock_exists(name):
    '''Returns True if lock of the given name exists'''
    return os.path.exists(name)


def sanitize_column(column_name):
    from frappe import _
    regex = re.compile("^.*[,'();].*")
    blacklisted_keywords = ['select', 'create', 'insert', 'delete', 'drop', 'update', 'case', 'and', 'or']

    def _raise_exception():
        frappe.throw(_("Invalid field name {0}").format(column_name), frappe.DataError)

    if 'ifnull' in column_name:
        if regex.match(column_name):
            # to avoid and, or
            if any(' {0} '.format(keyword) in column_name.split() for keyword in blacklisted_keywords):
                _raise_exception()

            # to avoid select, delete, drop, update and case
            elif any(keyword in column_name.split() for keyword in blacklisted_keywords):
                _raise_exception()

    elif regex.match(column_name):
        _raise_exception()


def cint(s):
    """Convert to integer"""
    try: num = int(float(s))
    except: num = 0
    return num


def make_filter_tuple(doctype, key, value):
    '''return a filter tuple like [doctype, key, operator, value]'''
    if isinstance(value, (list, tuple)):
        return [doctype, key, value[0], value[1]]
    else:
        return [doctype, key, "=", value]


number_format_info = {
    "#,###.##": (".", ",", 2),
    "#.###,##": (",", ".", 2),
    "# ###.##": (".", " ", 2),
    "# ###,##": (",", " ", 2),
    "#'###.##": (".", "'", 2),
    "#, ###.##": (".", ", ", 2),
    "#,##,###.##": (".", ",", 2),
    "#,###.###": (".", ",", 3),
    "#.###": ("", ".", 0),
    "#,###": ("", ",", 0)
}

operator_map = {
    # startswith
    "^": lambda a, b: (a or "").startswith(b),

    # in or not in a list
    "in": lambda a, b: operator.contains(b, a),
    "not in": lambda a, b: not operator.contains(b, a),

    # comparison operators
    "=": lambda a, b: operator.eq(a, b),
    "!=": lambda a, b: operator.ne(a, b),
    ">": lambda a, b: operator.gt(a, b),
    "<": lambda a, b: operator.lt(a, b),
    ">=": lambda a, b: operator.ge(a, b),
    "<=": lambda a, b: operator.le(a, b),
    "not None": lambda a, b: a and True or False,
    "None": lambda a, b: (not a) and True or False
}


def compare(val1, condition, val2):
    ret = False
    if condition in operator_map:
        ret = operator_map[condition](val1, val2)

    return ret


def format_time(txt):
    try:
        formatted_time = babel.dates.format_time(get_time(txt), locale=(frappe.local.lang or "").replace("-", "_"))
    except UnknownLocaleError:
        formatted_time = get_time(txt).strftime("%H:%M:%S")
    return formatted_time


def get_datetime_str(datetime_obj):
    if isinstance(datetime_obj, string_types):
        datetime_obj = get_datetime(datetime_obj)
    return datetime_obj.strftime(DATETIME_FORMAT)


def date_diff(string_ed_date, string_st_date):
    return (getdate(string_ed_date) - getdate(string_st_date)).days


def format_datetime(datetime_string, format_string=None):
    if not datetime_string:
        return

    datetime = get_datetime(datetime_string)
    if not format_string:
        format_string = get_user_format().replace("mm", "MM") + " HH:mm:ss"

    try:
        formatted_datetime = babel.dates.format_datetime(datetime, format_string,
                                                         locale=(frappe.local.lang or "").replace("-", "_"))
    except UnknownLocaleError:
        formatted_datetime = datetime.strftime('%Y-%m-%d %H:%M:%S')
    return formatted_datetime


def get_number_format_info(format):
    return number_format_info.get(format) or (".", ",", 2)


def fmt_money(amount, precision=None, currency=None):
    """
    Convert to string with commas for thousands, millions etc
    """
    number_format = frappe.db.get_default("number_format") or "#,###.##"
    if precision is None:
        precision = cint(frappe.db.get_default('currency_precision')) or None

    decimal_str, comma_str, number_format_precision = get_number_format_info(number_format)

    if precision is None:
        precision = number_format_precision

    # 40,000 -> 40,000.00
    # 40,000.00000 -> 40,000.00
    # 40,000.23000 -> 40,000.23

    if isinstance(amount, string_types):
        amount = flt(amount, precision)

    if decimal_str:
        decimals_after = str(round(amount % 1, precision))
        parts = decimals_after.split('.')
        parts = parts[1] if len(parts) > 1 else parts[0]
        decimals = parts
        if precision > 2:
            if len(decimals) < 3:
                if currency:
                    fraction = frappe.db.get_value("Currency", currency, "fraction_units", cache=True) or 100
                    precision = len(cstr(fraction)) - 1
                else:
                    precision = number_format_precision
            elif len(decimals) < precision:
                precision = len(decimals)

    amount = '%.*f' % (precision, round(flt(amount), precision))

    if amount.find('.') == -1:
        decimals = ''
    else:
        decimals = amount.split('.')[1]

    parts = []
    minus = ''
    if flt(amount) < 0:
        minus = '-'

    amount = cstr(abs(flt(amount))).split('.')[0]

    if len(amount) > 3:
        parts.append(amount[-3:])
        amount = amount[:-3]

        val = number_format == "#,##,###.##" and 2 or 3

        while len(amount) > val:
            parts.append(amount[-val:])
            amount = amount[:-val]

    parts.append(amount)

    parts.reverse()

    amount = comma_str.join(parts) + ((precision and decimal_str) and (decimal_str + decimals) or "")
    if amount != '0':
        amount = minus + amount

    if currency and frappe.get_global_default("hide_currency_symbol") != "Yes":
        symbol = frappe.db.get_value("Currency", currency, "symbol", cache=True) or currency
        amount = symbol + " " + amount

    return amount


# datetime functions
def formatdate(string_date=None, format_string=None):
    """
        Converts the given string date to :data:`user_format`
        User format specified in defaults

         Examples:

         * dd-mm-yyyy
         * mm-dd-yyyy
         * dd/mm/yyyy
    """

    if not string_date:
        return ''

    date = getdate(string_date)
    if not format_string:
        format_string = get_user_format()
    format_string = format_string.replace("mm", "MM")
    try:
        formatted_date = babel.dates.format_date(date, format_string,
                                                 locale=(frappe.local.lang or "").replace("-", "_"))
    except UnknownLocaleError:
        format_string = format_string.replace("MM", "%m").replace("dd", "%d").replace("yyyy", "%Y")
        formatted_date = date.strftime(format_string)
    return formatted_date


def get_user_format():
    if getattr(frappe.local, "user_format", None) is None:
        frappe.local.user_format = frappe.db.get_default("date_format")

    return frappe.local.user_format or "yyyy-mm-dd"


def nowdate():
    """return current date as yyyy-mm-dd"""
    return now_datetime().strftime(DATE_FORMAT)


def nowtime():
    """return current time in hh:mm"""
    return now_datetime().strftime(TIME_FORMAT)


def add_to_date(date, years=0, months=0, weeks=0, days=0, hours=0, minutes=0, seconds=0, as_string=False,
                as_datetime=False):
    """Adds `days` to the given date"""
    from dateutil.relativedelta import relativedelta

    if date == None:
        date = now_datetime()

    if hours:
        as_datetime = True

    if isinstance(date, string_types):
        as_string = True
        if " " in date:
            as_datetime = True
        date = parser.parse(date)

    date = date + relativedelta(years=years, months=months, weeks=weeks, days=days, hours=hours, minutes=minutes,
                                seconds=seconds)

    if as_string:
        if as_datetime:
            return date.strftime(DATETIME_FORMAT)
        else:
            return date.strftime(DATE_FORMAT)
    else:
        return date


def get_time(time_str):
    if isinstance(time_str, datetime.datetime):
        return time_str.time()
    elif isinstance(time_str, datetime.time):
        return time_str
    else:
        if isinstance(time_str, datetime.timedelta):
            time_str = str(time_str)
        return parser.parse(time_str).time()


def to_timedelta(time_str):
    if isinstance(time_str, string_types):
        t = parser.parse(time_str)
        return datetime.timedelta(hours=t.hour, minutes=t.minute, seconds=t.second, microseconds=t.microsecond)

    else:
        return time_str


def getdate(string_date=None):
    """
    Converts string date (yyyy-mm-dd) to datetime.date object
    """

    if not string_date:
        return get_datetime().date()
    if isinstance(string_date, datetime.datetime):
        return string_date.date()

    elif isinstance(string_date, datetime.date):
        return string_date

    # dateutil parser does not agree with dates like 0001-01-01
    if not string_date or string_date == "0001-01-01":
        return None
    return parser.parse(string_date).date()


def get_datetime(datetime_str=None):
    if not datetime_str or datetime_str == "0000-00-00 00:00:00.000000":
        return now_datetime()

    if isinstance(datetime_str, (datetime.datetime, datetime.timedelta)):
        return datetime_str

    elif isinstance(datetime_str, (list, tuple)):
        return datetime.datetime(datetime_str)

    elif isinstance(datetime_str, datetime.date):
        return datetime.datetime.combine(datetime_str, datetime.time())

    # dateutil parser does not agree with dates like 0001-01-01
    if not datetime_str or (datetime_str or "").startswith("0001-01-01"):
        return None

    try:
        return datetime.datetime.strptime(datetime_str, DATETIME_FORMAT)
    except ValueError:
        return parser.parse(datetime_str)


def now():
    """return current datetime as yyyy-mm-dd hh:mm:ss"""
    if frappe.flags.current_date:
        return getdate(frappe.flags.current_date).strftime(DATE_FORMAT) + " " + \
               now_datetime().strftime(TIME_FORMAT)
    else:
        return now_datetime().strftime(DATETIME_FORMAT)


def now_datetime():
    dt = convert_utc_to_user_timezone(datetime.datetime.utcnow())
    return dt.replace(tzinfo=None)


def _get_time_zone():
    return frappe.db.get_system_setting('time_zone') or 'Asia/Kolkata'  # Default to India ?!


def get_time_zone():
    if frappe.local.flags.in_test:
        return _get_time_zone()

    return frappe.cache().get_value("time_zone", _get_time_zone)


def convert_utc_to_user_timezone(utc_timestamp):
    from pytz import timezone, UnknownTimeZoneError
    utcnow = timezone('UTC').localize(utc_timestamp)
    try:
        return utcnow.astimezone(timezone(get_time_zone()))
    except UnknownTimeZoneError:
        return utcnow


def encode(obj, encoding="utf-8"):
    # if isinstance(obj, list):
    #     out = []
    #     for o in obj:
    #         if isinstance(o, text_type):
    #             out.append(o.encode(encoding))
    #         else:
    #             out.append(o)
    #     return out
    # elif isinstance(obj, text_type):
    #     return obj.encode(encoding)
    # else:
    return obj


def log(method, message=None):
    """log error in patch_log"""
    message = frappe.utils.cstr(message) + "\n" if message else ""
    message += frappe.get_traceback()

    if not (frappe.db and frappe.db._conn):
        frappe.connect()

    frappe.db.rollback()
    frappe.db.begin()

    d = frappe.new_doc("Error Log")
    d.method = method
    d.error = message
    d.insert(ignore_permissions=True)

    frappe.db.commit()

    return message



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
import frappe

# select doctypes that are accessed by the user (not read_only) first, so that the
# the validation message shows the user-facing doctype first.
# For example Journal Entry should be validated before GL Entry (which is an internal doctype)

dynamic_link_queries =  [
	"""select `tabDocField`.parent,
		`tabDocType`.read_only, `tabDocType`.in_create,
		`tabDocField`.fieldname, `tabDocField`.options
	from `tabDocField`, `tabDocType`
	where `tabDocField`.fieldtype='Dynamic Link' and
	`tabDocType`.`name`=`tabDocField`.parent
	order by `tabDocType`.read_only, `tabDocType`.in_create""",

	"""select `tabCustom Field`.dt as parent,
		`tabDocType`.read_only, `tabDocType`.in_create,
		`tabCustom Field`.fieldname, `tabCustom Field`.options
	from `tabCustom Field`, `tabDocType`
	where `tabCustom Field`.fieldtype='Dynamic Link' and
	`tabDocType`.`name`=`tabCustom Field`.dt
	order by `tabDocType`.read_only, `tabDocType`.in_create""",
]

def get_dynamic_link_map(for_delete=False):
	'''Build a map of all dynamically linked tables. For example,
		if Note is dynamically linked to ToDo, the function will return
		`{"Note": ["ToDo"], "Sales Invoice": ["Journal Entry Detail"]}`

	Note: Will not map single doctypes
	'''
	if getattr(frappe.local, 'dynamic_link_map', None)==None or frappe.flags.in_test:
		# Build from scratch
		dynamic_link_map = {}
		for df in get_dynamic_links():
			meta = frappe.get_meta(df.parent)
			if meta.issingle:
				# always check in Single DocTypes
				dynamic_link_map.setdefault(meta.name, []).append(df)
			else:
				links = frappe.db.sql_list("""select distinct {options} from `tab{parent}`""".format(**df))
				for doctype in links:
					dynamic_link_map.setdefault(doctype, []).append(df)

		frappe.local.dynamic_link_map = dynamic_link_map
	return frappe.local.dynamic_link_map

def get_dynamic_links():
	'''Return list of dynamic link fields as DocField.
	Uses cache if possible'''
	df = []
	for query in dynamic_link_queries:
		df += frappe.db.sql(query, as_dict=True)
	return df



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals

'''
File based locking utility
'''

import os
from time import time


class LockTimeoutError(Exception):
    pass


def create_lock(name):
    return False


def lock_exists(name):
    '''Returns True if lock of the given name exists'''
    return os.path.exists(get_lock_path(name))


def check_lock(path, timeout=600):
    return False


def delete_lock(name):
    return True


def get_lock_path(name):
    return ""



from __future__ import unicode_literals

import frappe
import os, sys
from frappe.database.db_manager import DbManager

expected_settings_10_2_earlier = {
	"innodb_file_format": "Barracuda",
	"innodb_file_per_table": "ON",
	"innodb_large_prefix": "ON",
	"character_set_server": "utf8mb4",
	"collation_server": "utf8mb4_unicode_ci"
}

expected_settings_10_3_later = {
	"character_set_server": "utf8mb4",
	"collation_server": "utf8mb4_unicode_ci"
}


def get_mariadb_versions():
	# MariaDB classifies their versions as Major (1st and 2nd number), and Minor (3rd number)
	# Example: Version 10.3.13 is Major Version = 10.3, Minor Version = 13
	mariadb_variables = frappe._dict(frappe.db.sql("""show variables"""))
	version_string = mariadb_variables.get('version').split('-')[0]
	versions = {}
	versions['major'] = version_string.split(
		'.')[0] + '.' + version_string.split('.')[1]
	versions['minor'] = version_string.split('.')[2]
	return versions


def setup_database(force, source_sql, verbose):
	frappe.local.session = frappe._dict({'user':'Administrator'})

	db_name = frappe.local.conf.db_name
	root_conn = get_root_connection(frappe.flags.root_login, frappe.flags.root_password)
	dbman = DbManager(root_conn)
	if force or (db_name not in dbman.get_database_list()):
		dbman.delete_user(db_name)
		dbman.drop_database(db_name)
	else:
		raise Exception("Database %s already exists" % (db_name,))

	dbman.create_user(db_name, frappe.conf.db_password)
	if verbose: print("Created user %s" % db_name)

	dbman.create_database(db_name)
	if verbose: print("Created database %s" % db_name)

	dbman.grant_all_privileges(db_name, db_name)
	dbman.flush_privileges()
	if verbose: print("Granted privileges to user %s and database %s" % (db_name, db_name))

	# close root connection
	root_conn.close()

	bootstrap_database(db_name, verbose, source_sql)

def setup_help_database(help_db_name):
	dbman = DbManager(get_root_connection(frappe.flags.root_login, frappe.flags.root_password))
	dbman.drop_database(help_db_name)

	# make database
	if not help_db_name in dbman.get_database_list():
		try:
			dbman.create_user(help_db_name, help_db_name)
		except Exception as e:
			# user already exists
			if e.args[0] != 1396: raise
		dbman.create_database(help_db_name)
		dbman.grant_all_privileges(help_db_name, help_db_name)
		dbman.flush_privileges()

def drop_user_and_database(db_name, root_login, root_password):
	frappe.local.db = get_root_connection(root_login, root_password)
	dbman = DbManager(frappe.local.db)
	dbman.delete_user(db_name)
	dbman.drop_database(db_name)

def bootstrap_database(db_name, verbose, source_sql=None):
	frappe.connect(db_name=db_name)
	if not check_database_settings():
		print('Database settings do not match expected values; stopping database setup.')
		sys.exit(1)

	import_db_from_sql(source_sql, verbose)
	if not 'tabDefaultValue' in frappe.db.get_tables():
		print('''Database not installed, this can due to lack of permission, or that the database name exists.
			Check your mysql root password, or use --force to reinstall''')
		sys.exit(1)

def import_db_from_sql(source_sql=None, verbose=False):
	if verbose: print("Starting database import...")
	db_name = frappe.conf.db_name
	if not source_sql:
		source_sql = os.path.join(os.path.dirname(__file__), 'framework_mariadb.sql')
	DbManager(frappe.local.db).restore_database(db_name, source_sql, db_name, frappe.conf.db_password)
	if verbose: print("Imported from database %s" % source_sql)


def check_database_settings():
	versions = get_mariadb_versions()
	if versions['major'] <= '10.2':
		expected_variables = expected_settings_10_2_earlier
	else:
		expected_variables = expected_settings_10_3_later

	mariadb_variables = frappe._dict(frappe.db.sql("""show variables"""))
	# Check each expected value vs. actuals:
	result = True
	for key, expected_value in expected_variables.items():
		if mariadb_variables.get(key) != expected_value:
			print("For key %s. Expected value %s, found value %s" %
				  (key, expected_value, mariadb_variables.get(key)))
			result = False
	if not result:
		site = frappe.local.site
		msg = ("Creation of your site - {x} failed because MariaDB is not properly {sep}"
			   "configured.  If using version 10.2.x or earlier, make sure you use the {sep}"
			   "the Barracuda storage engine. {sep}{sep}"
			   "Please verify the settings above in MariaDB's my.cnf.  Restart MariaDB.  And {sep}"
			   "then run `bench new-site {x}` again.{sep2}"
			   "").format(x=site, sep2="\n"*2, sep="\n")
		print_db_config(msg)
	return result


def get_root_connection(root_login, root_password):
	import getpass
	if not frappe.local.flags.root_connection:
		if not root_login:
			root_login = 'root'

		if not root_password:
			root_password = frappe.conf.get("root_password") or None

		if not root_password:
			root_password = getpass.getpass("MySQL root password: ")

		frappe.local.flags.root_connection = frappe.database.get_db(user=root_login, password=root_password)

	return frappe.local.flags.root_connection


def print_db_config(explanation):
	print("="*80)
	print(explanation)
	print("="*80)



from werkzeug.local import Local, release_local
import re, os, sys, importlib, inspect, json
from six import iteritems, binary_type, text_type, string_types
import frappe

local = Local()

# local-globals
db = local("db")
conf = local("conf")
form = form_dict = local("form_dict")
flags = local("flags")
error_log = local("error_log")
debug_log = local("debug_log")
message_log = local("message_log")

lang = local("lang")


class _dict(dict):
    """dict like object that exposes keys as attributes"""

    def __getattr__(self, key):
        ret = self.get(key)
        if not ret and key.startswith("__"):
            raise AttributeError()
        return ret

    def __setattr__(self, key, value):
        self[key] = value

    def __getstate__(self):
        return self

    def __setstate__(self, d):
        self.update(d)

    def update(self, d):
        """update and return self -- the missing dict feature in python"""
        super(_dict, self).update(d)
        return self

    def copy(self):
        return _dict(dict(self).copy())


class init_frappe:
    def __init__(self):
        '''If site==None, initialize it for empty site ('') to load common_site_config.json'''

    def __enter__(self):
        init()
        return local

    def __exit__(self, type, value, traceback):
        destroy()


def init(sites_path=None, new_site=False):
    """Initialize frappe for the current site. Reset thread locals `frappe.local`"""
    if getattr(local, "initialised", None):
        return

    local.error_log = []
    local.message_log = []
    local.debug_log = []
    local.realtime_log = []
    local.flags = _dict({
        "ran_schedulers": [],
        "currently_saving": [],
        "redirect_location": "",
        "in_install_db": False,
        "in_install_app": False,
        "in_import": False,
        "in_test": False,
        "mute_messages": False,
        "ignore_links": False,
        "mute_emails": False,
        "has_dataurl": False,
        "new_site": new_site
    })
    local.rollback_observers = []
    local.test_objects = {}

    local.site = "Nana"
    local.sites_path = sites_path
    local.site_path = os.path.abspath('.')
    local.request_ip = None
    local.response = _dict({"docs": []})
    local.task_id = None

    local.conf = _dict(get_site_config())
    local.lang = local.conf.lang or "en"
    local.lang_full_dict = None

    local.module_app = None
    local.app_modules = None
    local.system_settings = _dict()

    local.user = None
    local.user_perms = None
    local.session = None
    local.role_permissions = {}
    local.valid_columns = {}
    local.new_doc_templates = {}
    local.link_count = {}

    local.jenv = None
    local.jloader = None
    local.cache = {}
    local.document_cache = {}
    local.meta_cache = {}
    local.form_dict = _dict()
    local.session = _dict()

    # setup_module_map()

    local.initialised = True


def connect():
    """Connect to site database instance.

    :param site: If site is given, calls `frappe.init`.
    :param db_name: Optional. Will use from `site_config.json`."""
    from frappe.database import get_db
    init()

    local.db = get_db(user=getattr(local.conf, 'db_user', '') or local.conf.db_name)


def destroy():
    """Closes connection and releases werkzeug local."""
    if db:
        db.close()

    release_local(local)


def get_site_config():
    """Returns `site_config.json` combined with `sites/common_site_config.json`.
    `site_config` is a set of site wide settings like database name, password, email etc."""
    config = {}

    site_path = getattr(local, "site_path", None)

    if site_path:
        common_site_config = os.path.join(site_path, "common_site_config.json")
        if os.path.exists(common_site_config):
            config.update(get_file_json(common_site_config))

    # if site_path:
    #     site_config = os.path.join(site_path, "site_config.json")
    #     if os.path.exists(site_config):
    #         config.update(get_file_json(site_config))
    #     elif local.site and not local.flags.new_site:
    #         print("{0} does not exist".format(local.site))
    #         sys.exit(1)
    # raise IncorrectSitePath, "{0} does not exist".format(site_config)

    return _dict(config)


def get_conf(site=None):
    if hasattr(local, 'conf'):
        return local.conf

    else:
        # if no site, get from common_site_config.json
        with init_frappe(site):
            return local.conf


def get_file_json(path):
    """Read a file and return parsed JSON object."""
    with open(path, 'r') as f:
        return json.load(f)


def _(msg, lang=None):
    """Returns translated string in current lang, if exists."""
    if not hasattr(local, 'lang'):
        local.lang = lang or 'en'

    non_translated_msg = msg
    return non_translated_msg


def compare(val1, condition, val2):
    """Compare two values using `frappe.utils.compare`

    `condition` could be:
    - "^"
    - "in"
    - "not in"
    - "="
    - "!="
    - ">"
    - "<"
    - ">="
    - "<="
    - "not None"
    - "None"
    """
    import frappe.utils
    return frappe.utils.compare(val1, condition, val2)


# def html_to_js_template(path, content):
#     '''returns HTML template content as Javascript code, adding it to `frappe.templates`'''
#     return """frappe.templates["{key}"] = '{content}';\n""".format( \
#         key=path.rsplit("/", 1)[-1][:-5], content=scrub_html_template(content))
#
# def scrub_html_template(content):
#     '''Returns HTML content with removed whitespace and comments'''
#     # remove whitespace to a single space
#     content = re.sub("\s+", " ", content)
#
#     # strip comments
#     content =  re.sub("(<!--.*?-->)", "", content)
#
#     return content.replace("'", "\'")


def clear_cache(user=None, doctype=None):
    """Clear **User**, **DocType** or global cache.

    :param user: If user is given, only user cache is cleared.
    :param doctype: If doctype is given, only DocType cache is cleared."""
    pass


def safe_decode(param, encoding='utf-8'):
    try:
        param = param.decode(encoding)
    except Exception:
        pass
    return param


def as_unicode(text, encoding='utf-8'):
    '''Convert to unicode if required'''
    if isinstance(text, text_type):
        return text
    elif text == None:
        return ''
    elif isinstance(text, binary_type):
        return text_type(text, encoding)
    else:
        return text_type(text)


def get_meta(doctype, cached=True):
    """Get `frappe.model.meta.Meta` instance of given doctype name."""
    import frappe.model.meta
    return frappe.model.meta.get_meta(doctype, cached=cached)


def bold(text):
    return '<b>{0}</b>'.format(text)


def has_permission(doctype=None, ptype="read", doc=None, user=None, verbose=False, throw=False):
    """Raises `frappe.PermissionError` if not permitted.

    :param doctype: DocType for which permission is to be check.
    :param ptype: Permission type (`read`, `write`, `create`, `submit`, `cancel`, `amend`). Default: `read`.
    :param doc: [optional] Checks User permissions for given doc.
    :param user: [optional] Check for given user. Default: current user."""
    if not doctype and doc:
        doctype = doc.doctype

    out = True
    if throw and not out:
        if doc:
            frappe.throw(_("No permission for {0}").format(doc.doctype + " " + doc.name))
        else:
            frappe.throw(_("No permission for {0}").format(doctype))

    return out


def throw(msg, exc=0, title=None):
    """Throw execption and show message (`msgprint`).

    :param msg: Message.
    :param exc: Exception class. Default `frappe.ValidationError`"""
    msgprint(msg, raise_exception=exc, title=title, indicator='red')


def msgprint(msg, title=None, raise_exception=0, as_table=False, indicator=None, alert=False):
    """Print a message to the user (via HTTP response).
    Messages are sent in the `__server_messages` property in the
    response JSON and shown in a pop-up / modal.

    :param msg: Message.
    :param title: [optional] Message title.
    :param raise_exception: [optional] Raise given exception and show message.
    :param as_table: [optional] If `msg` is a list of lists, render as HTML table.
    """
    from frappe.utils import encode

    msg = safe_decode(msg)
    out = _dict(message=msg)

    def _raise_exception():
        if raise_exception:
            if flags.rollback_on_exception:
                db.rollback()
            import inspect

            if inspect.isclass(raise_exception) and issubclass(raise_exception, Exception):
                raise raise_exception(msg)
            else:
                raise ValidationError(msg)

    if flags.mute_messages:
        _raise_exception()
        return

    if as_table and type(msg) in (list, tuple):
        out.msg = '<table border="1px" style="border-collapse: collapse" cellpadding="2px">' + ''.join(
            ['<tr>' + ''.join(['<td>%s</td>' % c for c in r]) + '</tr>' for r in msg]) + '</table>'

    if flags.print_messages and out.msg:
        print("Message: " + repr(out.msg).encode("utf-8"))

    if title:
        out.title = title

    if not indicator and raise_exception:
        indicator = 'red'

    if indicator:
        out.indicator = indicator

    if alert:
        out.alert = 1

    message_log.append(json.dumps(out))

    if raise_exception and hasattr(raise_exception, '__name__'):
        local.response['exc_type'] = raise_exception.__name__

    _raise_exception()


def get_list(doctype, *args, **kwargs):
    """List database query via `frappe.model.db_query`. Will also check for permissions.

    :param doctype: DocType on which query is to be made.
    :param fields: List of fields or `*`.
    :param filters: List of filters (see example).
    :param order_by: Order By e.g. `modified desc`.
    :param limit_page_start: Start results at record #. Default 0.
    :param limit_page_length: No of records in the page. Default 20.

    Example usage:

        # simple dict filter
        frappe.get_list("ToDo", fields=["name", "description"], filters = {"owner":"test@example.com"})

        # filter as a list of lists
        frappe.get_list("ToDo", fields="*", filters = [["modified", ">", "2014-01-01"]])

        # filter as a list of dicts
        frappe.get_list("ToDo", fields="*", filters = {"description": ("like", "test%")})
    """
    import frappe.model.db_query
    return frappe.model.db_query.DatabaseQuery(doctype).execute(None, *args, **kwargs)


def get_all(doctype, *args, **kwargs):
    """List database query via `frappe.model.db_query`. Will **not** check for permissions.
    Parameters are same as `frappe.get_list`

    :param doctype: DocType on which query is to be made.
    :param fields: List of fields or `*`. Default is: `["name"]`.
    :param filters: List of filters (see example).
    :param order_by: Order By e.g. `modified desc`.
    :param limit_start: Start results at record #. Default 0.
    :param limit_page_length: No of records in the page. Default 20.

    Example usage:

        # simple dict filter
        frappe.get_all("ToDo", fields=["name", "description"], filters = {"owner":"test@example.com"})

        # filter as a list of lists
        frappe.get_all("ToDo", fields=["*"], filters = [["modified", ">", "2014-01-01"]])

        # filter as a list of dicts
        frappe.get_all("ToDo", fields=["*"], filters = {"description": ("like", "test%")})
    """
    kwargs["ignore_permissions"] = True
    if not "limit_page_length" in kwargs:
        kwargs["limit_page_length"] = 0
    return get_list(doctype, *args, **kwargs)


def get_value(*args, **kwargs):
    """Returns a document property or list of properties.

    Alias for `frappe.db.get_value`

    :param doctype: DocType name.
    :param filters: Filters like `{"x":"y"}` or name of the document. `None` if Single DocType.
    :param fieldname: Column name.
    :param ignore: Don't raise exception if table, column is missing.
    :param as_dict: Return values as dict.
    :param debug: Print query in error log.
    """
    return db.get_value(*args, **kwargs)


if sys.version_info.major == 2:
    class FileNotFoundError(Exception):
        pass
else:
    from builtins import FileNotFoundError


class ValidationError(Exception):
    http_status_code = 417


class AuthenticationError(Exception):
    http_status_code = 401


class SessionExpired(Exception):
    http_status_code = 401


class PermissionError(Exception):
    http_status_code = 403


class DoesNotExistError(ValidationError):
    http_status_code = 404


class NameError(Exception):
    http_status_code = 409


class OutgoingEmailError(Exception):
    http_status_code = 501


class SessionStopped(Exception):
    http_status_code = 503


class UnsupportedMediaType(Exception):
    http_status_code = 415


class RequestToken(Exception):
    http_status_code = 200


class Redirect(Exception):
    http_status_code = 301


class CSRFTokenError(Exception):
    http_status_code = 400


class ImproperDBConfigurationError(Exception):
    """
    Used when frappe detects that database or tables are not properly
    configured
    """

    def __init__(self, reason, msg=None):
        if not msg:
            msg = "MariaDb is not properly configured"
        super(ImproperDBConfigurationError, self).__init__(msg)
        self.reason = reason


def html_to_js_template(path, content):
    '''returns HTML template content as Javascript code, adding it to `frappe.templates`'''
    return """frappe.templates["{key}"] = '{content}';\n""".format( \
        key=path.rsplit("/", 1)[-1][:-5], content=scrub_html_template(content))


def scrub_html_template(content):
    '''Returns HTML content with removed whitespace and comments'''
    # remove whitespace to a single space
    content = re.sub("\s+", " ", content)

    # strip comments
    content = re.sub("(<!--.*?-->)", "", content)

    return content.replace("'", "\'")


def get_doc(*args, **kwargs):
    """Return a `frappe.model.document.Document` object of the given type and name.

    :param arg1: DocType name as string **or** document JSON.
    :param arg2: [optional] Document name as string.

    Examples:

        # insert a new document
        todo = frappe.get_doc({"doctype":"ToDo", "description": "test"})
        tood.insert()

        # open an existing document
        todo = frappe.get_doc("ToDo", "TD0001")

    """
    import frappe.model.document
    doc = frappe.model.document.get_doc(*args, **kwargs)

    return doc


def new_doc(doctype, parent_doc=None, parentfield=None, as_dict=False):
    """Returns a new document of the given DocType with defaults set.

    :param doctype: DocType of the new document.
    :param parent_doc: [optional] add to parent document.
    :param parentfield: [optional] add against this `parentfield`."""
    from frappe.model.create_new import get_new_doc
    return get_new_doc(doctype, parent_doc, parentfield, as_dict=as_dict)


def set_value(doctype, docname, fieldname, value=None):
    pass


# frappe Initialize
connect()


# Global
def get_defaults(user=None):
    globald = get_defaults_for()

    if not user:
        user = "Guest"

    if user:
        userd = {}
        userd.update(get_defaults_for(user))
        userd.update({"user": user, "owner": user})
        globald.update(userd)

    return globald


def set_global_default(key, value):
    set_default(key, value, "__default")


def add_global_default(key, value):
    add_default(key, value, "__default")


def get_global_default(key):
    d = get_defaults().get(key, None)

    value = isinstance(d, (list, tuple)) and d[0] or d
    return value


# Common

def get_module(modulename):
    """Returns a module object for given Python module name using `importlib.import_module`."""
    pass


def scrub(txt):
    """Returns sluggified string. e.g. `Sales Order` becomes `sales_order`."""
    return txt.replace(' ', '_').replace('-', '_').lower()


def set_default(key, value, parent, parenttype="__default"):
    """Override or add a default value.
    Adds default value in table `tabDefaultValue`.

    :param key: Default key.
    :param value: Default value.
    :param parent: Usually, **User** to whom the default belongs.
    :param parenttype: [optional] default is `__default`."""
    if frappe.db.sql('''
		select
			defkey
		from
			`tabDefaultValue`
		where
			defkey=%s and parent=%s
		for update''', (key, parent)):
        frappe.db.sql("""
			delete from
				`tabDefaultValue`
			where
				defkey=%s and parent=%s""", (key, parent))
    if value != None:
        add_default(key, value, parent)
    else:
        pass


def add_default(key, value, parent, parenttype=None):
    d = frappe.get_doc({
        "doctype": "DefaultValue",
        "parent": parent,
        "parenttype": parenttype or "__default",
        "parentfield": "system_defaults",
        "defkey": key,
        "defvalue": value
    })
    d.insert(ignore_permissions=True)


def clear_default(key=None, value=None, parent=None, name=None, parenttype=None):
    """Clear a default value by any of the given parameters and delete caches.

    :param key: Default key.
    :param value: Default value.
    :param parent: User name, or `__global`, `__default`.
    :param name: Default ID.
    :param parenttype: Clear defaults table for a particular type e.g. **User**.
    """
    conditions = []
    values = []

    if name:
        conditions.append("name=%s")
        values.append(name)

    else:
        if key:
            conditions.append("defkey=%s")
            values.append(key)

        if value:
            conditions.append("defvalue=%s")
            values.append(value)

        if parent:
            conditions.append("parent=%s")
            values.append(parent)

        if parenttype:
            conditions.append("parenttype=%s")
            values.append(parenttype)

    if not conditions:
        raise Exception("[clear_default] No key specified.")

    frappe.db.sql("""delete from tabDefaultValue where {0}""".format(" and ".join(conditions)),
                  tuple(values))


def get_defaults_for(parent="__default"):
    """get all defaults"""
    defaults = None

    if defaults == None:
        # sort descending because first default must get precedence
        res = frappe.db.sql("""select defkey, defvalue from `tabDefaultValue`
			where parent = %s order by creation""", (parent,), as_dict=1)

        defaults = frappe._dict({})
        for d in res:
            if d.defkey in defaults:
                # listify
                if not isinstance(defaults[d.defkey], list) and defaults[d.defkey] != d.defvalue:
                    defaults[d.defkey] = [defaults[d.defkey]]

                if d.defvalue not in defaults[d.defkey]:
                    defaults[d.defkey].append(d.defvalue)

            elif d.defvalue is not None:
                defaults[d.defkey] = d.defvalue

    return defaults


class DuplicateEntryError(NameError): pass


class DataError(ValidationError): pass


class UnknownDomainError(Exception): pass


class MappingMismatchError(ValidationError): pass


class InvalidStatusError(ValidationError): pass


class MandatoryError(ValidationError): pass


class InvalidSignatureError(ValidationError): pass


class RateLimitExceededError(ValidationError): pass


class CannotChangeConstantError(ValidationError): pass


class CharacterLengthExceededError(ValidationError): pass


class UpdateAfterSubmitError(ValidationError): pass


class LinkValidationError(ValidationError): pass


class CancelledLinkError(LinkValidationError): pass


class DocstatusTransitionError(ValidationError): pass


class TimestampMismatchError(ValidationError): pass


class EmptyTableError(ValidationError): pass


class LinkExistsError(ValidationError): pass


class InvalidEmailAddressError(ValidationError): pass


class TemplateNotFoundError(ValidationError): pass


class UniqueValidationError(ValidationError): pass


class AppNotInstalledError(ValidationError): pass


class ImplicitCommitError(ValidationError): pass


class RetryBackgroundJobError(Exception): pass


class DocumentLockedError(ValidationError): pass


class CircularLinkingError(ValidationError): pass


class SecurityException(Exception): pass


class InvalidColumnName(ValidationError): pass


class IncompatibleApp(ValidationError): pass


class InvalidDates(ValidationError): pass


class DataTooLongException(ValidationError): pass



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

# model __init__.py
from __future__ import unicode_literals
import frappe

data_fieldtypes = (
	'Currency',
	'Int',
	'Long Int',
	'Float',
	'Percent',
	'Check',
	'Small Text',
	'Long Text',
	'Code',
	'Text Editor',
	'Markdown Editor',
	'HTML Editor',
	'Date',
	'Datetime',
	'Time',
	'Text',
	'Data',
	'Link',
	'Dynamic Link',
	'Password',
	'Select',
	'Rating',
	'Read Only',
	'Attach',
	'Attach Image',
	'Signature',
	'Color',
	'Barcode',
	'Geolocation'
)

no_value_fields = ('Section Break', 'Column Break', 'HTML', 'Table', 'Table MultiSelect', 'Button', 'Image',
	'Fold', 'Heading')
display_fieldtypes = ('Section Break', 'Column Break', 'HTML', 'Button', 'Image', 'Fold', 'Heading')
numeric_fieldtypes = ('Currency', 'Int', 'Long Int', 'Float', 'Percent', 'Check')
default_fields = ('doctype','name','owner','creation','modified','modified_by',
	'parent','parentfield','parenttype','idx','docstatus')
optional_fields = ("_user_tags", "_comments", "_assign", "_liked_by", "_seen")
table_fields = ('Table', 'Table MultiSelect')
core_doctypes_list = ('DocType', 'DocField', 'DocPerm', 'User', 'Role', 'Has Role',
	'Page', 'Module Def', 'Print Format', 'Report', 'Customize Form',
	'Customize Form Field', 'Property Setter', 'Custom Field', 'Custom Script')

def copytables(srctype, src, srcfield, tartype, tar, tarfield, srcfields, tarfields=[]):
	if not tarfields:
		tarfields = srcfields
	l = []
	data = src.get(srcfield)
	for d in data:
		newrow = tar.append(tarfield)
		newrow.idx = d.idx

		for i in range(len(srcfields)):
			newrow.set(tarfields[i], d.get(srcfields[i]))

		l.append(newrow)
	return l



def db_exists(dt, dn):
	return frappe.db.exists(dt, dn)

def delete_fields(args_dict, delete=0):
	"""
		Delete a field.
		* Deletes record from `tabDocField`
		* If not single doctype: Drops column from table
		* If single, deletes record from `tabSingles`
		args_dict = { dt: [field names] }
	"""
	import frappe.utils
	for dt in args_dict:
		fields = args_dict[dt]
		if not fields:
			continue

		frappe.db.sql("""
			DELETE FROM `tabDocField`
			WHERE parent='%s' AND fieldname IN (%s)
		""" % (dt, ", ".join(["'{}'".format(f) for f in fields])))

		# Delete the data/column only if delete is specified
		if not delete:
			continue

		if frappe.db.get_value("DocType", dt, "issingle"):
			frappe.db.sql("""
				DELETE FROM `tabSingles`
				WHERE doctype='%s' AND field IN (%s)
			""" % (dt, ", ".join(["'{}'".format(f) for f in fields])))
		else:
			existing_fields = frappe.db.multisql({
					"mariadb": "DESC `tab%s`" % dt,
					"postgres": """
						SELECT
 							COLUMN_NAME
						FROM
 							information_schema.COLUMNS
						WHERE
 							TABLE_NAME = 'tab%s';
					""" % dt,
				})
			existing_fields = existing_fields and [e[0] for e in existing_fields] or []
			fields_need_to_delete = set(fields) & set(existing_fields)
			if not fields_need_to_delete:
				continue

			if frappe.db.db_type == 'mariadb':
				# mariadb implicitly commits before DDL, make it explicit
				frappe.db.commit()

			query = "ALTER TABLE `tab%s` " % dt + \
				", ".join(["DROP COLUMN `%s`" % f for f in fields_need_to_delete])
			frappe.db.sql(query)

		if frappe.db.db_type == 'postgres':
			# commit the results to db
			frappe.db.commit()



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals, print_function
from six.moves import range
import frappe
from frappe import _
from frappe.utils import cstr
from frappe import html_to_js_template
import re
from six import text_type

import io

STANDARD_FIELD_CONVERSION_MAP = {
	'name': 'Link',
	'owner': 'Data',
	'idx': 'Int',
	'creation': 'Data',
	'modified': 'Data',
	'modified_by': 'Data',
	'_user_tags': 'Data',
	'_liked_by': 'Data',
	'_comments': 'Text',
	'_assign': 'Text',
	'docstatus': 'Int'
}

"""
Model utilities, unclassified functions
"""

def set_default(doc, key):
	"""Set is_default property of given doc and unset all others filtered by given key."""
	if not doc.is_default:
		frappe.db.set(doc, "is_default", 1)

	frappe.db.sql("""update `tab%s` set `is_default`=0
		where `%s`=%s and name!=%s""" % (doc.doctype, key, "%s", "%s"),
		(doc.get(key), doc.name))

def set_field_property(filters, key, value):
	'''utility set a property in all fields of a particular type'''
	docs = [frappe.get_doc('DocType', d.parent) for d in \
		frappe.get_all("DocField", fields=['parent'], filters=filters)]

	for d in docs:
		d.get('fields', filters)[0].set(key, value)
		d.save()
		print('Updated {0}'.format(d.name))

	frappe.db.commit()

class InvalidIncludePath(frappe.ValidationError): pass

def render_include(content):
	'''render {% raw %}{% include "app/path/filename" %}{% endraw %} in js file'''

	content = cstr(content)

	# try 5 levels of includes
	for i in range(5):
		if "{% include" in content:
			paths = re.findall(r'''{% include\s['"](.*)['"]\s%}''', content)
			if not paths:
				frappe.throw(_('Invalid include path'), InvalidIncludePath)

			for path in paths:
				app, app_path = path.split('/', 1)
				with io.open(frappe.get_app_path(app, app_path), 'r', encoding = 'utf-8') as f:
					include = f.read()
					if path.endswith('.html'):
						include = html_to_js_template(path, include)

					content = re.sub(r'''{{% include\s['"]{0}['"]\s%}}'''.format(path), include, content)

		else:
			break

	return content

def get_fetch_values(doctype, fieldname, value):
	'''Returns fetch value dict for the given object

	:param doctype: Target doctype
	:param fieldname: Link fieldname selected
	:param value: Value selected
	'''
	out = {}
	meta = frappe.get_meta(doctype)
	link_df = meta.get_field(fieldname)
	for df in meta.get_fields_to_fetch(fieldname):
		# example shipping_address.gistin
		link_field, source_fieldname = df.fetch_from.split('.', 1)
		out[df.fieldname] = frappe.db.get_value(link_df.options, value, source_fieldname)

	return out



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
"""docfield utililtes"""

import frappe

def rename(doctype, fieldname, newname):
	"""rename docfield"""
	df = frappe.db.sql("""select * from tabDocField where parent=%s and fieldname=%s""",
		(doctype, fieldname), as_dict=1)
	if not df:
		return

	df = df[0]

	if frappe.db.get_value('DocType', doctype, 'issingle'):
		update_single(df, newname)
	else:
		update_table(df, newname)
		update_parent_field(df, newname)

def update_single(f, new):
	"""update in tabSingles"""
	frappe.db.begin()
	frappe.db.sql("""update tabSingles set field=%s where doctype=%s and field=%s""",
		(new, f['parent'], f['fieldname']))
	frappe.db.commit()

def update_table(f, new):
	"""update table"""
	query = get_change_column_query(f, new)
	if query:
		frappe.db.sql(query)

def update_parent_field(f, new):
	"""update 'parentfield' in tables"""
	if f['fieldtype'] in frappe.model.table_fields:
		frappe.db.begin()
		frappe.db.sql("""update `tab%s` set parentfield=%s where parentfield=%s""" \
			% (f['options'], '%s', '%s'), (new, f['fieldname']))
		frappe.db.commit()

def get_change_column_query(f, new):
	"""generate change fieldname query"""
	desc = frappe.db.sql("desc `tab%s`" % f['parent'])
	for d in desc:
		if d[0]== f['fieldname']:
			return 'alter table `tab%s` change `%s` `%s` %s' % \
				(f['parent'], f['fieldname'], new, d[1])

def supports_translation(fieldtype):
	return fieldtype in ["Data", "Select", "Text", "Small Text", "Text Editor"]


# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
import frappe
import datetime
from frappe.utils import formatdate, fmt_money, flt, cstr, cint, format_datetime, format_time
from frappe.model.meta import get_field_currency, get_field_precision
import re
from six import string_types


def format_value(value, df=None, doc=None, currency=None, translated=False):
    '''Format value based on given fieldtype, document reference, currency reference.
    If docfield info (df) is not given, it will try and guess based on the datatype of the value'''
    if isinstance(df, string_types):
        df = frappe._dict(fieldtype=df)

    if not df:
        df = frappe._dict()
        if isinstance(value, datetime.datetime):
            df.fieldtype = 'Datetime'
        elif isinstance(value, datetime.date):
            df.fieldtype = 'Date'
        elif isinstance(value, datetime.timedelta):
            df.fieldtype = 'Time'
        elif isinstance(value, int):
            df.fieldtype = 'Int'
        elif isinstance(value, float):
            df.fieldtype = 'Float'
        else:
            df.fieldtype = 'Data'

    elif (isinstance(df, dict)):
        # Convert dict to object if necessary
        df = frappe._dict(df)

    if value is None:
        value = ""
    elif translated:
        value = frappe._(value)

    if not df:
        return value

    elif df.get("fieldtype")=="Date":
        return formatdate(value)

    elif df.get("fieldtype")=="Datetime":
        return format_datetime(value)

    elif df.get("fieldtype")=="Time":
        return format_time(value)

    elif value==0 and df.get("fieldtype") in ("Int", "Float", "Currency", "Percent") and df.get("print_hide_if_no_value"):
        # this is required to show 0 as blank in table columns
        return ""

    elif df.get("fieldtype") == "Currency" or (df.get("fieldtype")=="Float" and (df.options or "").strip()):
        return fmt_money(value, precision=get_field_precision(df, doc),
                         currency=currency if currency else (get_field_currency(df, doc) if doc else None))

    elif df.get("fieldtype") == "Float":
        precision = get_field_precision(df, doc)

        # show 1.000000 as 1
        # options should not specified
        if not df.options and value is not None:
            temp = cstr(value).split(".")
            if len(temp)==1 or cint(temp[1])==0:
                precision = 0

        return fmt_money(value, precision=precision)

    elif df.get("fieldtype") == "Percent":
        return "{}%".format(flt(value, 2))

    elif df.get("fieldtype") in ("Text", "Small Text"):
        if not re.search("(\<br|\<div|\<p)", value):
            return frappe.safe_decode(value).replace("\n", "<br>")

    elif df.get("fieldtype") == "Markdown Editor":
        return  value

    elif df.get("fieldtype") == "Table MultiSelect":
        meta = frappe.get_meta(df.options)
        link_field = [df for df in meta.fields if df.fieldtype == 'Link'][0]
        values = [v.get(link_field.fieldname, 'asdf') for v in value]
        return ', '.join(values)

    return value



from __future__ import unicode_literals

import frappe
import warnings

import pymysql
from pymysql.times import TimeDelta
from pymysql.constants 	import ER, FIELD_TYPE
from pymysql.converters import conversions

from frappe.utils import get_datetime, cstr
from markdown2 import UnicodeWithAttrs
from frappe.database.database import Database
from six import PY2, binary_type, text_type, string_types
from frappe.database.mariadb.schema import MariaDBTable


class MariaDBDatabase(Database):
	ProgrammingError = pymysql.err.ProgrammingError
	TableMissingError = pymysql.err.ProgrammingError
	OperationalError = pymysql.err.OperationalError
	InternalError = pymysql.err.InternalError
	SQLError = pymysql.err.ProgrammingError
	DataError = pymysql.err.DataError
	REGEX_CHARACTER = 'regexp'

	def setup_type_map(self):
		self.db_type = 'mariadb'
		self.type_map = {
			'Currency':		('decimal', '18,6'),
			'Int':			('int', '11'),
			'Long Int':		('bigint', '20'),
			'Float':		('decimal', '18,6'),
			'Percent':		('decimal', '18,6'),
			'Check':		('int', '1'),
			'Small Text':	('text', ''),
			'Long Text':	('longtext', ''),
			'Code':			('longtext', ''),
			'Text Editor':	('longtext', ''),
			'Markdown Editor':	('longtext', ''),
			'HTML Editor':	('longtext', ''),
			'Date':			('date', ''),
			'Datetime':		('datetime', '6'),
			'Time':			('time', '6'),
			'Text':			('text', ''),
			'Data':			('varchar', self.VARCHAR_LEN),
			'Link':			('varchar', self.VARCHAR_LEN),
			'Dynamic Link':	('varchar', self.VARCHAR_LEN),
			'Password':		('varchar', self.VARCHAR_LEN),
			'Select':		('varchar', self.VARCHAR_LEN),
			'Rating':		('int', '1'),
			'Read Only':	('varchar', self.VARCHAR_LEN),
			'Attach':		('text', ''),
			'Attach Image':	('text', ''),
			'Signature':	('longtext', ''),
			'Color':		('varchar', self.VARCHAR_LEN),
			'Barcode':		('longtext', ''),
			'Geolocation':	('longtext', '')
		}

	def get_connection(self):
		warnings.filterwarnings('ignore', category=pymysql.Warning)
		usessl = 0
		if frappe.conf.db_ssl_ca and frappe.conf.db_ssl_cert and frappe.conf.db_ssl_key:
			usessl = 1
			ssl_params = {
				'ca':frappe.conf.db_ssl_ca,
				'cert':frappe.conf.db_ssl_cert,
				'key':frappe.conf.db_ssl_key
			}

		conversions.update({
			FIELD_TYPE.NEWDECIMAL: float,
			FIELD_TYPE.DATETIME: get_datetime,
			UnicodeWithAttrs: conversions[text_type]
		})

		if PY2:
			conversions.update({
				TimeDelta: conversions[binary_type]
			})
		self.user = getattr(frappe.conf, "db_user", None) or frappe.conf.db_name

		if usessl:
			conn = pymysql.connect(self.host, getattr(frappe.conf, "db_user", None) or self.user or '', self.password or '', database=frappe.conf.db_name,
				port=self.port, charset='utf8mb4', use_unicode = True, ssl=ssl_params,
				conv = conversions, local_infile = frappe.conf.local_infile)
		else:
			conn = pymysql.connect(self.host, self.user or '', self.password or '',
				port=self.port, charset='utf8mb4', use_unicode = True, conv = conversions,
				local_infile = frappe.conf.local_infile)

		# MYSQL_OPTION_MULTI_STATEMENTS_OFF = 1
		# # self._conn.set_server_option(MYSQL_OPTION_MULTI_STATEMENTS_OFF)
		print(str(frappe.conf.db_name))
		if self.user != 'root':
			conn.select_db(frappe.conf.db_name)

		return conn

	def get_database_size(self):
		''''Returns database size in MB'''
		db_size = self.sql('''
			SELECT `table_schema` as `database_name`,
			SUM(`data_length` + `index_length`) / 1024 / 1024 AS `database_size`
			FROM information_schema.tables WHERE `table_schema` = %s GROUP BY `table_schema`
			''', self.db_name, as_dict=True)

		return db_size[0].get('database_size')

	@staticmethod
	def escape(s, percent=True):
		"""Excape quotes and percent in given string."""
		# pymysql expects unicode argument to escape_string with Python 3
		s = frappe.as_unicode(pymysql.escape_string(frappe.as_unicode(s)), "utf-8").replace("`", "\\`")

		# NOTE separating % escape, because % escape should only be done when using LIKE operator
		# or when you use python format string to generate query that already has a %s
		# for example: sql("select name from `tabUser` where name=%s and {0}".format(conditions), something)
		# defaulting it to True, as this is the most frequent use case
		# ideally we shouldn't have to use ESCAPE and strive to pass values via the values argument of sql
		if percent:
			s = s.replace("%", "%%")

		return "'" + s + "'"

	# column type
	@staticmethod
	def is_type_number(code):
		return code == pymysql.NUMBER

	@staticmethod
	def is_type_datetime(code):
		return code in (pymysql.DATE, pymysql.DATETIME)

	# exception types
	@staticmethod
	def is_deadlocked(e):
		return e.args[0] == ER.LOCK_DEADLOCK

	@staticmethod
	def is_timedout(e):
		return e.args[0] == ER.LOCK_WAIT_TIMEOUT

	@staticmethod
	def is_table_missing(e):
		return e.args[0] == ER.NO_SUCH_TABLE

	@staticmethod
	def is_missing_column(e):
		return e.args[0] == ER.BAD_FIELD_ERROR

	@staticmethod
	def is_duplicate_fieldname(e):
		return e.args[0] == ER.DUP_FIELDNAME

	@staticmethod
	def is_duplicate_entry(e):
		return e.args[0] == ER.DUP_ENTRY

	@staticmethod
	def is_access_denied( e):
		return e.args[0] == ER.ACCESS_DENIED_ERROR

	@staticmethod
	def cant_drop_field_or_key(e):
		return e.args[0] == ER.CANT_DROP_FIELD_OR_KEY

	@staticmethod
	def is_syntax_error(e):
		return e.args[0] == ER.PARSE_ERROR

	@staticmethod
	def is_data_too_long(e):
		return e.args[0] == ER.DATA_TOO_LONG

	def is_primary_key_violation(self, e):
		return self.is_duplicate_entry(e) and 'PRIMARY' in cstr(e.args[1])

	def is_unique_key_violation(self, e):
		return self.is_duplicate_entry(e) and 'Duplicate' in cstr(e.args[1])


	def create_auth_table(self):
		self.sql_ddl("""create table if not exists `__Auth` (
				`doctype` VARCHAR(140) NOT NULL,
				`name` VARCHAR(255) NOT NULL,
				`fieldname` VARCHAR(140) NOT NULL,
				`password` VARCHAR(255) NOT NULL,
				`encrypted` INT(1) NOT NULL DEFAULT 0,
				PRIMARY KEY (`doctype`, `name`, `fieldname`)
			) ENGINE=InnoDB ROW_FORMAT=COMPRESSED CHARACTER SET=utf8mb4 COLLATE=utf8mb4_unicode_ci""")

	def create_global_search_table(self):
		if not '__global_search' in self.get_tables():
			self.sql('''create table __global_search(
				doctype varchar(100),
				name varchar({0}),
				title varchar({0}),
				content text,
				fulltext(content),
				route varchar({0}),
				published int(1) not null default 0,
				unique `doctype_name` (doctype, name))
				COLLATE=utf8mb4_unicode_ci
				ENGINE=MyISAM
				CHARACTER SET=utf8mb4'''.format(self.VARCHAR_LEN))

	def create_user_settings_table(self):
		self.sql_ddl("""create table if not exists __UserSettings (
			`user` VARCHAR(180) NOT NULL,
			`doctype` VARCHAR(180) NOT NULL,
			`data` TEXT,
			UNIQUE(user, doctype)
			) ENGINE=InnoDB DEFAULT CHARSET=utf8""")

	def create_help_table(self):
		self.sql('''create table help(
				path varchar(255),
				content text,
				title text,
				intro text,
				full_path text,
				fulltext(title),
				fulltext(content),
				index (path))
				COLLATE=utf8mb4_unicode_ci
				ENGINE=MyISAM
				CHARACTER SET=utf8mb4''')

	@staticmethod
	def get_on_duplicate_update(key=None):
		return 'ON DUPLICATE key UPDATE '

	def get_table_columns_description(self, table_name):
		"""Returns list of column and its description"""
		return self.sql('''select
			column_name as 'name',
			column_type as 'type',
			column_default as 'default',
			column_key = 'MUL' as 'index',
			column_key = 'UNI' as 'unique'
			from information_schema.columns
			where table_name = '{table_name}' '''.format(table_name=table_name), as_dict=1)

	def has_index(self, table_name, index_name):
		return self.sql("""SHOW INDEX FROM `{table_name}`
			WHERE Key_name='{index_name}'""".format(
				table_name=table_name,
				index_name=index_name
			))

	def add_index(self, doctype, fields, index_name=None):
		"""Creates an index with given fields if not already created.
		Index name will be `fieldname1_fieldname2_index`"""
		index_name = index_name or self.get_index_name(fields)
		table_name = 'tab' + doctype
		if not self.has_index(table_name, index_name):
			self.commit()
			self.sql("""ALTER TABLE `%s`
				ADD INDEX `%s`(%s)""" % (table_name, index_name, ", ".join(fields)))

	def add_unique(self, doctype, fields, constraint_name=None):
		if isinstance(fields, string_types):
			fields = [fields]
		if not constraint_name:
			constraint_name = "unique_" + "_".join(fields)

		if not self.sql("""select CONSTRAINT_NAME from information_schema.TABLE_CONSTRAINTS
			where table_name=%s and constraint_type='UNIQUE' and CONSTRAINT_NAME=%s""",
			('tab' + doctype, constraint_name)):
				self.commit()
				self.sql("""alter table `tab%s`
					add unique `%s`(%s)""" % (doctype, constraint_name, ", ".join(fields)))

	def updatedb(self, doctype, meta=None):
		"""
		Syncs a `DocType` to the table
		* creates if required
		* updates columns
		* updates indices
		"""
		res = self.sql("select issingle from `tabDocType` where name=%s", (doctype,))
		if not res:
			raise Exception('Wrong doctype {0} in updatedb'.format(doctype))

		if not res[0][0]:
			db_table = MariaDBTable(doctype, meta)
			db_table.validate()

			self.commit()
			db_table.sync()
			self.begin()

	def get_database_list(self, target):
		return [d[0] for d in self.sql("SHOW DATABASES;")]



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
import frappe
from frappe import _
from frappe.utils import now_datetime, cint, cstr
import re
from six import string_types


def set_new_name(doc):
	"""
	Sets the `name` property for the document based on various rules.

	1. If amended doc, set suffix.
	2. If `autoname` method is declared, then call it.
	3. If `autoname` property is set in the DocType (`meta`), then build it using the `autoname` property.
	4. If no rule defined, use hash.

	:param doc: Document to be named.
	"""

	doc.run_method("before_naming")

	autoname = frappe.get_meta(doc.doctype).autoname or ""

	if autoname.lower() != "prompt" and not frappe.flags.in_import:
		doc.name = None

	if getattr(doc, "amended_from", None):
		_set_amended_name(doc)
		return

	elif getattr(doc.meta, "issingle", False):
		doc.name = doc.doctype

	else:
		doc.run_method("autoname")

	if not doc.name and autoname:
		set_name_from_naming_options(autoname, doc)

	# if the autoname option is 'field:' and no name was derived, we need to
	# notify
	if autoname.startswith('field:') and not doc.name:
		fieldname = autoname[6:]
		frappe.throw(_("{0} is required").format(doc.meta.get_label(fieldname)))

	# at this point, we fall back to name generation with the hash option
	if not doc.name or autoname == 'hash':
		doc.name = make_autoname('hash', doc.doctype)

	doc.name = validate_name(
		doc.doctype,
		doc.name,
		frappe.get_meta(doc.doctype).get_field("name_case")
	)


def set_name_from_naming_options(autoname, doc):
	"""
	Get a name based on the autoname field option
	"""

	_autoname = autoname.lower()

	if _autoname.startswith('field:'):
		doc.name = _field_autoname(autoname, doc)
	elif _autoname.startswith('naming_series:'):
		set_name_by_naming_series(doc)
	elif _autoname.startswith('prompt'):
		_prompt_autoname(autoname, doc)
	elif _autoname.startswith('format:'):
		doc.name = _format_autoname(autoname, doc)
	elif '#' in autoname:
		doc.name = make_autoname(autoname, doc=doc)

def set_name_by_naming_series(doc):
	"""Sets name by the `naming_series` property"""
	if not doc.naming_series:
		doc.naming_series = get_default_naming_series(doc.doctype)

	if not doc.naming_series:
		frappe.throw(frappe._("Naming Series mandatory"))

	doc.name = make_autoname(doc.naming_series+'.#####', '', doc)

def make_autoname(key='', doctype='', doc=''):
	"""
	Creates an autoname from the given key:

	**Autoname rules:**

		 * The key is separated by '.'
		 * '####' represents a series. The string before this part becomes the prefix:
			Example: ABC.#### creates a series ABC0001, ABC0002 etc
		 * 'MM' represents the current month
		 * 'YY' and 'YYYY' represent the current year


   *Example:*

		 * DE/./.YY./.MM./.##### will create a series like
		   DE/09/01/0001 where 09 is the year, 01 is the month and 0001 is the series
	"""
	if key == "hash":
		return frappe.generate_hash(doctype, 10)

	if "#" not in key:
		key = key + ".#####"
	elif "." not in key:
		frappe.throw(_("Invalid naming series (. missing)") + (_(" for {0}").format(doctype) if doctype else ""))

	parts = key.split('.')
	n = parse_naming_series(parts, doctype, doc)
	return n


def parse_naming_series(parts, doctype='', doc=''):
	n = ''
	if isinstance(parts, string_types):
		parts = parts.split('.')
	series_set = False
	today = now_datetime()
	for e in parts:
		part = ''
		if e.startswith('#'):
			if not series_set:
				digits = len(e)
				part = getseries(n, digits)
				series_set = True
		elif e == 'YY':
			part = today.strftime('%y')
		elif e == 'MM':
			part = today.strftime('%m')
		elif e == 'DD':
			part = today.strftime("%d")
		elif e == 'YYYY':
			part = today.strftime('%Y')
		elif e == 'FY':
			part = frappe.defaults.get_user_default("fiscal_year")
		elif e.startswith('{') and doc:
			e = e.replace('{', '').replace('}', '')
			part = doc.get(e)
		elif doc and doc.get(e):
			part = doc.get(e)
		else:
			part = e

		if isinstance(part, string_types):
			n += part

	return n


def getseries(key, digits):
	# series created ?
	current = frappe.db.sql("SELECT `current` FROM `tabSeries` WHERE `name`=%s FOR UPDATE", (key,))
	if current and current[0][0] is not None:
		current = current[0][0]
		# yes, update it
		frappe.db.sql("UPDATE `tabSeries` SET `current` = `current` + 1 WHERE `name`=%s", (key,))
		current = cint(current) + 1
	else:
		# no, create it
		frappe.db.sql("INSERT INTO `tabSeries` (`name`, `current`) VALUES (%s, 1)", (key,))
		current = 1
	return ('%0'+str(digits)+'d') % current


def revert_series_if_last(key, name):
	if ".#" in key:
		prefix, hashes = key.rsplit(".", 1)
		if "#" not in hashes:
			return
	else:
		prefix = key

	if '.' in prefix:
		prefix = parse_naming_series(prefix.split('.'))

	count = cint(name.replace(prefix, ""))
	current = frappe.db.sql("SELECT `current` FROM `tabSeries` WHERE `name`=%s FOR UPDATE", (prefix,))

	if current and current[0][0]==count:
		frappe.db.sql("UPDATE `tabSeries` SET `current` = `current` - 1 WHERE `name`=%s", prefix)


def get_default_naming_series(doctype):
	"""get default value for `naming_series` property"""
	naming_series = frappe.get_meta(doctype).get_field("naming_series").options or ""
	if naming_series:
		naming_series = naming_series.split("\n")
		return naming_series[0] or naming_series[1]
	else:
		return None


def validate_name(doctype, name, case=None, merge=False):
	if not name:
		return 'No Name Specified for %s' % doctype
	if name.startswith('New '+doctype):
		frappe.throw(_('There were some errors setting the name, please contact the administrator'), frappe.NameError)
	if case == 'Title Case':
		name = name.title()
	if case == 'UPPER CASE':
		name = name.upper()
	name = name.strip()

	if not frappe.get_meta(doctype).get("issingle") and (doctype == name) and (name != "DocType"):
		frappe.throw(_("Name of {0} cannot be {1}").format(doctype, name), frappe.NameError)

	special_characters = "<>"
	if re.findall("[{0}]+".format(special_characters), name):
		message = ", ".join("'{0}'".format(c) for c in special_characters)
		frappe.throw(_("Name cannot contain special characters like {0}").format(message), frappe.NameError)

	return name


def append_number_if_name_exists(doctype, value, fieldname='name', separator='-', filters=None):
	if not filters:
		filters = dict()
	filters.update({fieldname: value})
	exists = frappe.db.exists(doctype, filters)

	regex = '^{value}{separator}\d+$'.format(value=re.escape(value), separator=separator)

	if exists:
		last = frappe.db.sql("""SELECT `{fieldname}` FROM `tab{doctype}`
			WHERE `{fieldname}` {regex_character} %s
			ORDER BY length({fieldname}) DESC,
			`{fieldname}` DESC LIMIT 1""".format(
				doctype=doctype,
				fieldname=fieldname,
				regex_character=frappe.db.REGEX_CHARACTER),
			regex)

		if last:
			count = str(cint(last[0][0].rsplit(separator, 1)[1]) + 1)
		else:
			count = "1"

		value = "{0}{1}{2}".format(value, separator, count)

	return value


def _set_amended_name(doc):
	am_id = 1
	am_prefix = doc.amended_from
	if frappe.db.get_value(doc.doctype, doc.amended_from, "amended_from"):
		am_id = cint(doc.amended_from.split('-')[-1]) + 1
		am_prefix = '-'.join(doc.amended_from.split('-')[:-1])  # except the last hyphen

	doc.name = am_prefix + '-' + str(am_id)
	return doc.name


def _field_autoname(autoname, doc, skip_slicing=None):
	"""
	Generate a name using `DocType` field. This is called when the doctype's
	`autoname` field starts with 'field:'
	"""
	fieldname = autoname if skip_slicing else autoname[6:]
	name = (cstr(doc.get(fieldname)) or '').strip()
	return name


def _prompt_autoname(autoname, doc):
	"""
	Generate a name using Prompt option. This simply means the user will have to set the name manually.
	This is called when the doctype's `autoname` field starts with 'prompt'.
	"""
	# set from __newname in save.py
	if not doc.name:
		frappe.throw(_("Name not set via prompt"))

def _format_autoname(autoname, doc):
	"""
	Generate autoname by replacing all instances of braced params (fields, date params ('DD', 'MM', 'YY'), series)
	Independent of remaining string or separators.

	Example pattern: 'format:LOG-{MM}-{fieldname1}-{fieldname2}-{#####}'
	"""

	first_colon_index = autoname.find(':')
	autoname_value = autoname[first_colon_index + 1:]

	def get_param_value_for_match(match):
		param = match.group()
		# trim braces
		trimmed_param = param[1:-1]
		return parse_naming_series([trimmed_param], doc=doc)

	# Replace braced params with their parsed value
	name = re.sub(r'(\{[\w | #]+\})', get_param_value_for_match, autoname_value)

	return name



from __future__ import unicode_literals, print_function
import redis
from rq import Connection, Queue, Worker
from rq.logutils import setup_loghandlers
from frappe.utils import cstr
from collections import defaultdict
import frappe
import os, socket, time
from frappe import _
from six import string_types

# imports - third-party imports

default_timeout = 300
queue_timeout = {
	'background': 2500,
	'long': 1500,
	'default': 300,
	'short': 300,
	'firebase_long' : 1500,
	'firebase_short' : 300,
	'auto_assign_short' : 300,
	'auto_assign_long' : 1500,
	'third_party_sync_short' : 300,
	'third_party_sync_long' : 1500
}

redis_connection = None

def enqueue(method, queue='default', timeout=None, event=None,
	is_async=True, job_name=None, now=False, enqueue_after_commit=False, **kwargs):
	'''
		Enqueue method to be executed using a background worker

		:param method: method string or method object
		:param queue: should be either long, default or short
		:param timeout: should be set according to the functions
		:param event: this is passed to enable clearing of jobs from queues
		:param is_async: if is_async=False, the method is executed immediately, else via a worker
		:param job_name: can be used to name an enqueue call, which can be used to prevent duplicate calls
		:param now: if now=True, the method is executed via frappe.call
		:param kwargs: keyword arguments to be passed to the method
	'''
	# To handle older implementations
	is_async = kwargs.pop('async', is_async)

	if now or frappe.flags.in_migrate:
		return frappe.call(method, **kwargs)

	q = get_queue(queue, is_async=is_async)
	if not timeout:
		timeout = queue_timeout.get(queue) or 300
	queue_args = {
		"site": frappe.local.site,
		"user": frappe.session.user,
		"method": method,
		"event": event,
		"job_name": job_name or cstr(method),
		"is_async": is_async,
		"kwargs": kwargs
	}
	if enqueue_after_commit:
		if not frappe.flags.enqueue_after_commit:
			frappe.flags.enqueue_after_commit = []

		frappe.flags.enqueue_after_commit.append({
			"queue": queue,
			"is_async": is_async,
			"timeout": timeout,
			"queue_args":queue_args
		})
		return frappe.flags.enqueue_after_commit
	else:
		return q.enqueue_call(execute_job, timeout=timeout,
			kwargs=queue_args)

def enqueue_doc(doctype, name=None, method=None, queue='default', timeout=300,
	now=False, **kwargs):
	'''Enqueue a method to be run on a document'''
	enqueue('frappe.utils.background_jobs.run_doc_method', doctype=doctype, name=name,
		doc_method=method, queue=queue, timeout=timeout, now=now, **kwargs)

def run_doc_method(doctype, name, doc_method, **kwargs):
	getattr(frappe.get_doc(doctype, name), doc_method)(**kwargs)

def execute_job(site, method, event, job_name, kwargs, user=None, is_async=True, retry=0):
	'''Executes job in a worker, performs commit/rollback and logs if there is any error'''
	from frappe.utils import log

	if is_async:
		frappe.connect(site)
		if os.environ.get('CI'):
			frappe.flags.in_test = True

		if user:
			frappe.set_user(user)

	if isinstance(method, string_types):
		method_name = method
		method = frappe.get_attr(method)
	else:
		method_name = cstr(method.__name__)

	try:
		method(**kwargs)

	except (frappe.db.InternalError, frappe.RetryBackgroundJobError) as e:
		frappe.db.rollback()

		if (retry < 5 and
			(isinstance(e, frappe.RetryBackgroundJobError) or
				(frappe.db.is_deadlocked(e) or frappe.db.is_timedout(e)))):
			# retry the job if
			# 1213 = deadlock
			# 1205 = lock wait timeout
			# or RetryBackgroundJobError is explicitly raised
			frappe.destroy()
			time.sleep(retry+1)

			return execute_job(site, method, event, job_name, kwargs,
				is_async=is_async, retry=retry+1)

		else:
			log(method_name, message=repr(locals()))
			raise

	except:
		frappe.db.rollback()
		log(method_name, message=repr(locals()))
		raise

	else:
		frappe.db.commit()

	finally:
		if is_async:
			frappe.destroy()

def start_worker(queue=None, quiet = False):
	'''Wrapper to start rq worker. Connects to redis and monitors these queues.'''
	with frappe.init_site():
		# empty init is required to get redis_queue from common_site_config.json
		redis_connection = get_redis_conn()

	if os.environ.get('CI'):
		setup_loghandlers('ERROR')

	with Connection(redis_connection):
		queues = get_queue_list(queue)
		logging_level = "INFO"
		if quiet:
			logging_level = "WARNING"
		Worker(queues, name=get_worker_name(queue)).work(logging_level = logging_level)

def get_worker_name(queue):
	'''When limiting worker to a specific queue, also append queue name to default worker name'''
	name = None

	if queue:
		# hostname.pid is the default worker name
		name = '{hostname}.{pid}.{queue}'.format(
			hostname=socket.gethostname(),
			pid=os.getpid(),
			queue=queue)

	return name

def get_jobs(site=None, queue=None, key='method'):
	'''Gets jobs per queue or per site or both'''
	jobs_per_site = defaultdict(list)
	for queue in get_queue_list(queue):
		q = get_queue(queue)

		for job in q.jobs:
			if job.kwargs.get('site'):
				if site is None:
					# get jobs for all sites
					jobs_per_site[job.kwargs['site']].append(job.kwargs[key])

				elif job.kwargs['site'] == site:
					# get jobs only for given site
					jobs_per_site[site].append(job.kwargs[key])

			else:
				print('No site found in job', job.__dict__)

	return jobs_per_site

def get_queue_list(queue_list=None):
	'''Defines possible queues. Also wraps a given queue in a list after validating.'''
	default_queue_list = list(queue_timeout)
	if queue_list:
		if isinstance(queue_list, string_types):
			queue_list = [queue_list]

		for queue in queue_list:
			validate_queue(queue, default_queue_list)

		return queue_list

	else:
		return default_queue_list

def get_queue(queue, is_async=True):
	'''Returns a Queue object tied to a redis connection'''
	validate_queue(queue)

	kwargs = {
		'connection': get_redis_conn(),
		'async': is_async
	}

	return Queue(queue, **kwargs)

def validate_queue(queue, default_queue_list=None):
	if not default_queue_list:
		default_queue_list = list(queue_timeout)

	if queue not in default_queue_list:
		frappe.throw(_("Queue should be one of {0}").format(', '.join(default_queue_list)))

def get_redis_conn():
	if not hasattr(frappe.local, 'conf'):
		raise Exception('You need to call frappe.init')

	elif not frappe.local.conf.redis_queue:
		raise Exception('redis_queue missing in common_site_config.json')

	global redis_connection

	if not redis_connection:
		redis_connection = redis.from_url(frappe.local.conf.redis_queue)

	return redis_connection

def enqueue_test_job():
	enqueue('frappe.utils.background_jobs.test_job', s=100)

def test_job(s):
	import time
	print('sleeping...')
	time.sleep(s)



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

# Tree (Hierarchical) Nested Set Model (nsm)
#
# To use the nested set model,
# use the following pattern
# 1. name your parent field as "parent_item_group" if not have a property nsm_parent_field as your field name in the document class
# 2. have a field called "old_parent" in your fields list - this identifies whether the parent has been changed
# 3. call update_nsm(doc_obj) in the on_upate method

# ------------------------------------------
from __future__ import unicode_literals

import frappe
from frappe import _
from frappe.model.document import Document
from frappe.utils import now

class NestedSetRecursionError(frappe.ValidationError): pass
class NestedSetMultipleRootsError(frappe.ValidationError): pass
class NestedSetChildExistsError(frappe.ValidationError): pass
class NestedSetInvalidMergeError(frappe.ValidationError): pass

# called in the on_update method
def update_nsm(doc):
	# get fields, data from the DocType
	opf = 'old_parent'
	pf = "parent_" + frappe.scrub(doc.doctype)

	if hasattr(doc,'nsm_parent_field'):
		pf = doc.nsm_parent_field
	if hasattr(doc,'nsm_oldparent_field'):
		opf = doc.nsm_oldparent_field

	p, op = doc.get(pf) or None, doc.get(opf) or None

	# has parent changed (?) or parent is None (root)
	if not doc.lft and not doc.rgt:
		update_add_node(doc, p or '', pf)
	elif op != p:
		update_move_node(doc, pf)

	# set old parent
	doc.set(opf, p)
	frappe.db.set_value(doc.doctype, doc.name, opf, p or '', update_modified=False)

	doc.reload()

def update_add_node(doc, parent, parent_field):
	"""
		insert a new node
	"""

	n = now()

	doctype = doc.doctype
	name = doc.name

	# get the last sibling of the parent
	if parent:
		left, right = frappe.db.sql("select lft, rgt from `tab{0}` where name=%s"
			.format(doctype), parent)[0]
		validate_loop(doc.doctype, doc.name, left, right)
	else: # root
		right = frappe.db.sql("""
			SELECT COALESCE(MAX(rgt), 0) + 1 FROM `tab{0}`
			WHERE COALESCE(`{1}`, '') = ''
		""".format(doctype, parent_field))[0][0]
	right = right or 1

	# update all on the right
	frappe.db.sql("update `tab{0}` set rgt = rgt+2, modified=%s where rgt >= %s"
		.format(doctype), (n, right))
	frappe.db.sql("update `tab{0}` set lft = lft+2, modified=%s where lft >= %s"
		.format(doctype), (n, right))

	# update index of new node
	if frappe.db.sql("select * from `tab{0}` where lft=%s or rgt=%s".format(doctype), (right, right+1)):
		frappe.msgprint(_("Nested set error. Please contact the Administrator."))
		raise Exception

	frappe.db.sql("update `tab{0}` set lft=%s, rgt=%s, modified=%s where name=%s".format(doctype),
		(right,right+1, n, name))
	return right


def update_move_node(doc, parent_field):
	n = now()
	parent = doc.get(parent_field)

	if parent:
		new_parent = frappe.db.sql("""select lft, rgt from `tab{0}`
			where name = %s""".format(doc.doctype), parent, as_dict=1)[0]

		validate_loop(doc.doctype, doc.name, new_parent.lft, new_parent.rgt)

	# move to dark side
	frappe.db.sql("""update `tab{0}` set lft = -lft, rgt = -rgt, modified=%s
		where lft >= %s and rgt <= %s""".format(doc.doctype), (n, doc.lft, doc.rgt))

	# shift left
	diff = doc.rgt - doc.lft + 1
	frappe.db.sql("""update `tab{0}` set lft = lft -%s, rgt = rgt - %s, modified=%s
		where lft > %s""".format(doc.doctype), (diff, diff, n, doc.rgt))

	# shift left rgts of ancestors whose only rgts must shift
	frappe.db.sql("""update `tab{0}` set rgt = rgt - %s, modified=%s
		where lft < %s and rgt > %s""".format(doc.doctype), (diff, n, doc.lft, doc.rgt))

	if parent:
		new_parent = frappe.db.sql("""select lft, rgt from `tab%s`
			where name = %s""" % (doc.doctype, '%s'), parent, as_dict=1)[0]


		# set parent lft, rgt
		frappe.db.sql("""update `tab{0}` set rgt = rgt + %s, modified=%s
			where name = %s""".format(doc.doctype), (diff, n, parent))

		# shift right at new parent
		frappe.db.sql("""update `tab{0}` set lft = lft + %s, rgt = rgt + %s, modified=%s
			where lft > %s""".format(doc.doctype), (diff, diff, n, new_parent.rgt))

		# shift right rgts of ancestors whose only rgts must shift
		frappe.db.sql("""update `tab{0}` set rgt = rgt + %s, modified=%s
			where lft < %s and rgt > %s""".format(doc.doctype),
			(diff, n, new_parent.lft, new_parent.rgt))


		new_diff = new_parent.rgt - doc.lft
	else:
		# new root
		max_rgt = frappe.db.sql("""select max(rgt) from `tab{0}`""".format(doc.doctype))[0][0]
		new_diff = max_rgt + 1 - doc.lft

	# bring back from dark side
	frappe.db.sql("""update `tab{0}` set lft = -lft + %s, rgt = -rgt + %s, modified=%s
		where lft < 0""".format(doc.doctype), (new_diff, new_diff, n))

def rebuild_tree(doctype, parent_field):
	"""
		call rebuild_node for all root nodes
	"""
	# get all roots
	frappe.db.auto_commit_on_many_writes = 1

	right = 1
	result = frappe.db.sql("SELECT name FROM `tab%s` WHERE `%s`='' or `%s` IS NULL ORDER BY name ASC" % (doctype, parent_field, parent_field))
	for r in result:
		right = rebuild_node(doctype, r[0], right, parent_field)

	frappe.db.auto_commit_on_many_writes = 0

def rebuild_node(doctype, parent, left, parent_field):
	"""
		reset lft, rgt and recursive call for all children
	"""
	from frappe.utils import now
	n = now()

	# the right value of this node is the left value + 1
	right = left+1

	# get all children of this node
	result = frappe.db.sql("SELECT name FROM `tab{0}` WHERE `{1}`=%s"
		.format(doctype, parent_field), (parent))
	for r in result:
		right = rebuild_node(doctype, r[0], right, parent_field)

	# we've got the left value, and now that we've processed
	# the children of this node we also know the right value
	frappe.db.sql("""UPDATE `tab{0}` SET lft=%s, rgt=%s, modified=%s
		WHERE name=%s""".format(doctype), (left,right,n,parent))

	#return the right value of this node + 1
	return right+1


def validate_loop(doctype, name, lft, rgt):
	"""check if item not an ancestor (loop)"""
	if name in frappe.db.sql_list("""select name from `tab{0}` where lft <= %s and rgt >= %s"""
		 .format(doctype), (lft, rgt)):
		frappe.throw(_("Item cannot be added to its own descendents"), NestedSetRecursionError)

class NestedSet(Document):
	def __setup__(self):
		if self.meta.get("nsm_parent_field"):
			self.nsm_parent_field = self.meta.nsm_parent_field

	def on_update(self):
		update_nsm(self)
		self.validate_ledger()

	def on_trash(self, allow_root_deletion=False):
		if not getattr(self, 'nsm_parent_field', None):
			self.nsm_parent_field = frappe.scrub(self.doctype) + "_parent"

		parent = self.get(self.nsm_parent_field)
		if not parent and not allow_root_deletion:
			frappe.throw(_("Root {0} cannot be deleted").format(_(self.doctype)))

		# cannot delete non-empty group
		self.validate_if_child_exists()

		self.set(self.nsm_parent_field, "")

		try:
			update_nsm(self)
		except frappe.DoesNotExistError:
			if self.flags.on_rollback:
				pass
				frappe.message_log.pop()
			else:
				raise

	def validate_if_child_exists(self):
		has_children = frappe.db.sql("""select count(name) from `tab{doctype}`
			where `{nsm_parent_field}`=%s""".format(doctype=self.doctype, nsm_parent_field=self.nsm_parent_field),
			(self.name,))[0][0]
		if has_children:
			frappe.throw(_("Cannot delete {0} as it has child nodes").format(self.name), NestedSetChildExistsError)

	def before_rename(self, olddn, newdn, merge=False, group_fname="is_group"):
		if merge and hasattr(self, group_fname):
			is_group = frappe.db.get_value(self.doctype, newdn, group_fname)
			if self.get(group_fname) != is_group:
				frappe.throw(_("Merging is only possible between Group-to-Group or Leaf Node-to-Leaf Node"), NestedSetInvalidMergeError)

	def after_rename(self, olddn, newdn, merge=False):
		if not self.nsm_parent_field:
			parent_field = "parent_" + self.doctype.replace(" ", "_").lower()
		else:
			parent_field = self.nsm_parent_field

		# set old_parent for children
		frappe.db.sql("update `tab{0}` set old_parent=%s where {1}=%s"
			.format(self.doctype, parent_field), (newdn, newdn))

		if merge:
			rebuild_tree(self.doctype, parent_field)

	def validate_one_root(self):
		if not self.get(self.nsm_parent_field):
			if self.get_root_node_count() > 1:
				frappe.throw(_("""Multiple root nodes not allowed."""), NestedSetMultipleRootsError)

	def get_root_node_count(self):
		return frappe.db.count(self.doctype, {
			self.nsm_parent_field: ''
		})

	def validate_ledger(self, group_identifier="is_group"):
		if hasattr(self, group_identifier) and not bool(self.get(group_identifier)):
			if frappe.db.sql("""select name from `tab{0}` where {1}=%s and docstatus!=2"""
				.format(self.doctype, self.nsm_parent_field), (self.name)):
				frappe.throw(_("{0} {1} cannot be a leaf node as it has children").format(_(self.doctype), self.name))

	def get_ancestors(self):
		return get_ancestors_of(self.doctype, self.name)

def get_root_of(doctype):
	"""Get root element of a DocType with a tree structure"""
	result = frappe.db.sql("""select t1.name from `tab{0}` t1 where
		(select count(*) from `tab{1}` t2 where
			t2.lft < t1.lft and t2.rgt > t1.rgt) = 0
		and t1.rgt > t1.lft""".format(doctype, doctype))
	return result[0][0] if result else None

def get_ancestors_of(doctype, name, order_by="lft desc", limit=None):
	"""Get ancestor elements of a DocType with a tree structure"""
	lft, rgt = frappe.db.get_value(doctype, name, ["lft", "rgt"])

	result = [d["name"] for d in frappe.db.get_all(doctype, {"lft": ["<", lft], "rgt": [">", rgt]},
		"name", order_by=order_by, limit_page_length=limit)]

	return result or []

def get_descendants_of(doctype, name, order_by="lft desc", limit=None,
	ignore_permissions=False):
	'''Return descendants of the current record'''
	lft, rgt = frappe.db.get_value(doctype, name, ['lft', 'rgt'])

	result = [d["name"] for d in frappe.db.get_list(doctype, {"lft": [">", lft], "rgt": ["<", rgt]},
		"name", order_by=order_by, limit_page_length=limit, ignore_permissions=ignore_permissions)]

	return result or []



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
import os
from six import string_types, integer_types
import shutil

import frappe
import frappe.defaults
import frappe.model.meta
from frappe import _
from frappe.model.dynamic_links import get_dynamic_link_map
from frappe.model.naming import revert_series_if_last
from frappe import FileNotFoundError

doctypes_to_skip = (
"Communication", "ToDo", "DocShare", "Email Unsubscribe", "Activity Log", "File", "Version", "Document Follow",
"Comment", "View Log")


def delete_doc(doctype=None, name=None, force=0, ignore_doctypes=None, for_reload=False,
               ignore_permissions=False, flags=None, ignore_on_trash=False, ignore_missing=True):
    """
        Deletes a doc(dt, dn) and validates if it is not submitted and not linked in a live record
    """
    if not ignore_doctypes: ignore_doctypes = []

    # get from form
    if not doctype:
        doctype = frappe.form_dict.get('dt')
        name = frappe.form_dict.get('dn')

    names = name
    if isinstance(name, string_types) or isinstance(name, integer_types):
        names = [name]

    for name in names or []:

        # already deleted..?
        if not frappe.db.exists(doctype, name):
            if not ignore_missing:
                raise frappe.DoesNotExistError
            else:
                return False

        doc = None
        if doctype == "DocType":
            if for_reload:

                try:
                    doc = frappe.get_doc(doctype, name)
                except frappe.DoesNotExistError:
                    pass
                else:
                    doc.run_method("before_reload")

            else:
                doc = frappe.get_doc(doctype, name)

                update_flags(doc, flags, ignore_permissions)
                check_permission_and_not_submitted(doc)

                frappe.db.sql("delete from `tabCustom Field` where dt = %s", name)
                frappe.db.sql("delete from `tabCustom Script` where dt = %s", name)
                frappe.db.sql("delete from `tabProperty Setter` where doc_type = %s", name)
                frappe.db.sql("delete from `tabReport` where ref_doctype=%s", name)
                frappe.db.sql("delete from `tabCustom DocPerm` where parent=%s", name)
                frappe.db.sql("delete from `__global_search` where doctype=%s", name)

            delete_from_table(doctype, name, ignore_doctypes, None)

            if not (frappe.flags.in_migrate or frappe.flags.in_install or frappe.flags.in_test):
                try:
                    delete_controllers(name, doc.module)
                except (FileNotFoundError, OSError):
                    # in case a doctype doesnt have any controller code
                    pass

        else:
            doc = frappe.get_doc(doctype, name)

            if not for_reload:
                update_flags(doc, flags, ignore_permissions)
                check_permission_and_not_submitted(doc)

                if not ignore_on_trash:
                    doc.run_method("on_trash")
                    doc.flags.in_delete = True
                    doc.run_method('on_change')

                frappe.enqueue('frappe.model.delete_doc.delete_dynamic_links', doctype=doc.doctype, name=doc.name,
                               is_async=False if frappe.flags.in_test else True)

                # check if links exist
                if not force:
                    check_if_doc_is_linked(doc)
                    check_if_doc_is_dynamically_linked(doc)

            update_naming_series(doc)
            delete_from_table(doctype, name, ignore_doctypes, doc)
            doc.run_method("after_delete")

        if doc and not for_reload:
            add_to_deleted_document(doc)
            if not frappe.flags.in_patch:
                try:
                    doc.notify_update()
                    insert_feed(doc)
                except ImportError:
                    pass


def add_to_deleted_document(doc):
    '''Add this document to Deleted Document table. Called after delete'''
    if doc.doctype != 'Deleted Document' and frappe.flags.in_install != 'frappe':
        frappe.get_doc(dict(
            doctype='Deleted Document',
            deleted_doctype=doc.doctype,
            deleted_name=doc.name,
            data=doc.as_json(),
            owner="Administrator"
        )).db_insert()


def update_naming_series(doc):
    if doc.meta.autoname:
        if doc.meta.autoname.startswith("naming_series:") \
                and getattr(doc, "naming_series", None):
            revert_series_if_last(doc.naming_series, doc.name)

        elif doc.meta.autoname.split(":")[0] not in ("Prompt", "field", "hash"):
            revert_series_if_last(doc.meta.autoname, doc.name)


def delete_from_table(doctype, name, ignore_doctypes, doc):
    if doctype != "DocType" and doctype == name:
        frappe.db.sql("delete from `tabSingles` where `doctype`=%s", name)
    else:
        frappe.db.sql("delete from `tab{0}` where `name`=%s".format(doctype), name)

    # get child tables
    if doc:
        tables = [d.options for d in doc.meta.get_table_fields()]

    else:
        def get_table_fields(field_doctype):
            return [r[0] for r in frappe.get_all(field_doctype,
                                                 fields='options',
                                                 filters={
                                                     'fieldtype': ['in', frappe.model.table_fields],
                                                     'parent': doctype
                                                 },
                                                 as_list=1
                                                 )]

        tables = get_table_fields("DocField")
        if not frappe.flags.in_install == "frappe":
            tables += get_table_fields("Custom Field")

    # delete from child tables
    for t in list(set(tables)):
        if t not in ignore_doctypes:
            frappe.db.sql("delete from `tab%s` where parenttype=%s and parent = %s" % (t, '%s', '%s'), (doctype, name))


def update_flags(doc, flags=None, ignore_permissions=False):
    if ignore_permissions:
        if not flags: flags = {}
        flags["ignore_permissions"] = ignore_permissions

    if flags:
        doc.flags.update(flags)


def check_permission_and_not_submitted(doc):
    # permission
    if (not doc.flags.ignore_permissions
            and frappe.session.user != "Administrator"
            and (
                    not doc.has_permission("delete")
                    or (doc.doctype == "DocType" and not doc.custom))):
        frappe.msgprint(_("User not allowed to delete {0}: {1}")
                        .format(doc.doctype, doc.name), raise_exception=frappe.PermissionError)

    # check if submitted
    if doc.docstatus == 1:
        frappe.msgprint(_("{0} {1}: Submitted Record cannot be deleted.").format(_(doc.doctype), doc.name),
                        raise_exception=True)


def get_link_fields(doctype):
    # get link fields from tabDocField
    if not frappe.flags.link_fields:
        frappe.flags.link_fields = {}

    if not doctype in frappe.flags.link_fields:
        link_fields = frappe.db.sql("""\
			select parent, fieldname,
				(select issingle from tabDocType dt
				where dt.name = df.parent) as issingle
			from tabDocField df
			where
				df.options=%s and df.fieldtype='Link'""", (doctype,), as_dict=1)

        # get link fields from tabCustom Field
        custom_link_fields = frappe.db.sql("""\
			select dt as parent, fieldname,
				(select issingle from tabDocType dt
				where dt.name = df.dt) as issingle
			from `tabCustom Field` df
			where
				df.options=%s and df.fieldtype='Link'""", (doctype,), as_dict=1)

        # add custom link fields list to link fields list
        link_fields += custom_link_fields

        # remove fields whose options have been changed using property setter
        property_setter_link_fields = frappe.db.sql("""\
			select ps.doc_type as parent, ps.field_name as fieldname,
				(select issingle from tabDocType dt
				where dt.name = ps.doc_type) as issingle
			from `tabProperty Setter` ps
			where
				ps.property_type='options' and
				ps.field_name is not null and
				ps.value=%s""", (doctype,), as_dict=1)

        link_fields += property_setter_link_fields

        frappe.flags.link_fields[doctype] = link_fields

    return frappe.flags.link_fields[doctype]


def check_if_doc_is_linked(doc, method="Delete"):
    """
        Raises excption if the given doc(dt, dn) is linked in another record.
    """

    link_fields = get_link_fields(doc.doctype)
    link_fields = [[lf['parent'], lf['fieldname'], lf['issingle']] for lf in link_fields]

    for link_dt, link_field, issingle in link_fields:
        if not issingle:
            for item in frappe.db.get_values(link_dt, {link_field: doc.name},
                                             ["name", "parent", "parenttype", "docstatus"], as_dict=True):
                linked_doctype = item.parenttype if item.parent else link_dt

                ignore_linked_doctypes = doc.get('ignore_linked_doctypes') or []

                if linked_doctype in doctypes_to_skip or (
                        linked_doctype in ignore_linked_doctypes and method == 'Cancel'):
                    # don't check for communication and todo!
                    continue

                if not item:
                    continue
                elif method != "Delete" and (method != "Cancel" or item.docstatus != 1):
                    # don't raise exception if not
                    # linked to a non-cancelled doc when deleting or to a submitted doc when cancelling
                    continue
                elif link_dt == doc.doctype and (item.parent or item.name) == doc.name:
                    # don't raise exception if not
                    # linked to same item or doc having same name as the item
                    continue
                else:
                    reference_docname = item.parent or item.name
                    raise_link_exists_exception(doc, linked_doctype, reference_docname)

        else:
            if frappe.db.get_value(link_dt, None, link_field) == doc.name:
                raise_link_exists_exception(doc, link_dt, link_dt)


def check_if_doc_is_dynamically_linked(doc, method="Delete"):
    '''Raise `frappe.LinkExistsError` if the document is dynamically linked'''
    for df in get_dynamic_link_map().get(doc.doctype, []):

        ignore_linked_doctypes = doc.get('ignore_linked_doctypes') or []

        if df.parent in doctypes_to_skip or (df.parent in ignore_linked_doctypes and method == 'Cancel'):
            # don't check for communication and todo!
            continue

        meta = frappe.get_meta(df.parent)
        if meta.issingle:
            # dynamic link in single doc
            refdoc = frappe.db.get_singles_dict(df.parent)
            if (refdoc.get(df.options) == doc.doctype
                    and refdoc.get(df.fieldname) == doc.name
                    and ((method == "Delete" and refdoc.docstatus < 2)
                         or (method == "Cancel" and refdoc.docstatus == 1))
            ):
                # raise exception only if
                # linked to an non-cancelled doc when deleting
                # or linked to a submitted doc when cancelling
                raise_link_exists_exception(doc, df.parent, df.parent)
        else:
            # dynamic link in table
            df["table"] = ", `parent`, `parenttype`, `idx`" if meta.istable else ""
            for refdoc in frappe.db.sql("""select `name`, `docstatus` {table} from `tab{parent}` where
				{options}=%s and {fieldname}=%s""".format(**df), (doc.doctype, doc.name), as_dict=True):

                if ((method == "Delete" and refdoc.docstatus < 2) or (method == "Cancel" and refdoc.docstatus == 1)):
                    # raise exception only if
                    # linked to an non-cancelled doc when deleting
                    # or linked to a submitted doc when cancelling

                    reference_doctype = refdoc.parenttype if meta.istable else df.parent
                    reference_docname = refdoc.parent if meta.istable else refdoc.name
                    at_position = "at Row: {0}".format(refdoc.idx) if meta.istable else ""

                    raise_link_exists_exception(doc, reference_doctype, reference_docname, at_position)


def raise_link_exists_exception(doc, reference_doctype, reference_docname, row=''):
    doc_link = '<a href="#Form/{0}/{1}">{1}</a>'.format(doc.doctype, doc.name)
    reference_link = '<a href="#Form/{0}/{1}">{1}</a>'.format(reference_doctype, reference_docname)

    # hack to display Single doctype only once in message
    if reference_doctype == reference_docname:
        reference_doctype = ''

    frappe.throw(_('Cannot delete or cancel because {0} {1} is linked with {2} {3} {4}')
                 .format(doc.doctype, doc_link, reference_doctype, reference_link, row), frappe.LinkExistsError)


def delete_dynamic_links(doctype, name):
    delete_references('ToDo', doctype, name, 'reference_type')
    delete_references('Email Unsubscribe', doctype, name)
    delete_references('DocShare', doctype, name, 'share_doctype', 'share_name')
    delete_references('Version', doctype, name, 'ref_doctype', 'docname')
    delete_references('Comment', doctype, name)
    delete_references('View Log', doctype, name)
    delete_references('Document Follow', doctype, name, 'ref_doctype', 'ref_docname')

    # unlink communications
    clear_timeline_references(doctype, name)
    clear_references('Communication', doctype, name)

    clear_references('Activity Log', doctype, name)
    clear_references('Activity Log', doctype, name, 'timeline_doctype', 'timeline_name')


def delete_references(doctype, reference_doctype, reference_name,
                      reference_doctype_field='reference_doctype', reference_name_field='reference_name'):
    frappe.db.sql('''delete from `tab{0}`
		where {1}=%s and {2}=%s'''.format(doctype, reference_doctype_field, reference_name_field),  # nosec
                  (reference_doctype, reference_name))


def clear_references(doctype, reference_doctype, reference_name,
                     reference_doctype_field='reference_doctype', reference_name_field='reference_name'):
    frappe.db.sql('''update
			`tab{0}`
		set
			{1}=NULL, {2}=NULL
		where
			{1}=%s and {2}=%s'''.format(doctype, reference_doctype_field, reference_name_field),  # nosec
                  (reference_doctype, reference_name))


def clear_timeline_references(link_doctype, link_name):
    frappe.db.sql("""delete from `tabCommunication Link`
		where `tabCommunication Link`.link_doctype='{0}' and `tabCommunication Link`.link_name='{1}'""".format(
        link_doctype, link_name))  # nosec


def insert_feed(doc):
    if frappe.flags.in_install or frappe.flags.in_import or getattr(doc, "no_feed_on_delete", False):
        return

    frappe.get_doc({
        "doctype": "Comment",
        "comment_type": "Deleted",
        "reference_doctype": doc.doctype,
        "subject": "{0} {1}".format(_(doc.doctype), doc.name),
        "full_name": doc.owner
    }).insert(ignore_permissions=True)


def delete_controllers(doctype, module):
    pass



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals

import frappe
from six import iteritems

ignore_doctypes = ("DocType", "Print Format", "Role", "Module Def", "Communication",
	"ToDo")

def notify_link_count(doctype, name):
	'''updates link count for given document'''
	if hasattr(frappe.local, 'link_count'):
		if (doctype, name) in frappe.local.link_count:
			frappe.local.link_count[(doctype, name)] += 1
		else:
			frappe.local.link_count[(doctype, name)] = 1

def flush_local_link_count():
	'''flush from local before ending request'''
	if not getattr(frappe.local, 'link_count', None):
		return

	link_count = frappe.cache().get_value('_link_count')
	if not link_count:
		link_count = {}

		for key, value in frappe.local.link_count.items():
			if key in link_count:
				link_count[key] += frappe.local.link_count[key]
			else:
				link_count[key] = frappe.local.link_count[key]

	frappe.cache().set_value('_link_count', link_count)

def update_link_count():
	'''increment link count in the `idx` column for the given document'''
	link_count = frappe.cache().get_value('_link_count')

	if link_count:
		for key, count in iteritems(link_count):
			if key[0] not in ignore_doctypes:
				try:
					frappe.db.sql('update `tab{0}` set idx = idx + {1} where name=%s'.format(key[0], count),
						key[1], auto_commit=1)
				except Exception as e:
					if not frappe.db.is_table_missing(e): # table not found, single
						raise e
	# reset the count
	frappe.cache().delete_value('_link_count')



from __future__ import unicode_literals

import re
import frappe
import psycopg2
import psycopg2.extensions
from six import string_types
from frappe.utils import cstr
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from frappe.database.database import Database
from frappe.database.postgres.schema import PostgresTable

# cast decimals as floats
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    'DEC2FLOAT',
    lambda value, curs: float(value) if value is not None else None)

psycopg2.extensions.register_type(DEC2FLOAT)

class PostgresDatabase(Database):
	ProgrammingError = psycopg2.ProgrammingError
	TableMissingError = psycopg2.ProgrammingError
	OperationalError = psycopg2.OperationalError
	InternalError = psycopg2.InternalError
	SQLError = psycopg2.ProgrammingError
	DataError = psycopg2.DataError
	InterfaceError = psycopg2.InterfaceError
	REGEX_CHARACTER = '~'

	def setup_type_map(self):
		self.db_type = 'postgres'
		self.type_map = {
			'Currency':		('decimal', '18,6'),
			'Int':			('bigint', None),
			'Long Int':		('bigint', None),
			'Float':		('decimal', '18,6'),
			'Percent':		('decimal', '18,6'),
			'Check':		('smallint', None),
			'Small Text':	('text', ''),
			'Long Text':	('text', ''),
			'Code':			('text', ''),
			'Text Editor':	('text', ''),
			'Markdown Editor':	('text', ''),
			'HTML Editor':	('text', ''),
			'Date':			('date', ''),
			'Datetime':		('timestamp', None),
			'Time':			('time', '6'),
			'Text':			('text', ''),
			'Data':			('varchar', self.VARCHAR_LEN),
			'Link':			('varchar', self.VARCHAR_LEN),
			'Dynamic Link':	('varchar', self.VARCHAR_LEN),
			'Password':		('varchar', self.VARCHAR_LEN),
			'Select':		('varchar', self.VARCHAR_LEN),
			'Rating':		('smallint', None),
			'Read Only':	('varchar', self.VARCHAR_LEN),
			'Attach':		('text', ''),
			'Attach Image':	('text', ''),
			'Signature':	('text', ''),
			'Color':		('varchar', self.VARCHAR_LEN),
			'Barcode':		('text', ''),
			'Geolocation':	('text', '')
		}

	def get_connection(self):
		# warnings.filterwarnings('ignore', category=psycopg2.Warning)
		conn = psycopg2.connect("host='{}' dbname='{}' user='{}' password='{}' port={}".format(
			self.host, self.user, self.user, self.password, self.port
		))
		conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT) # TODO: Remove this

		return conn

	def escape(self, s, percent=True):
		"""Excape quotes and percent in given string."""
		if isinstance(s, bytes):
			s = s.decode('utf-8')

		if percent:
			s = s.replace("%", "%%")

		s = s.encode('utf-8')

		return str(psycopg2.extensions.QuotedString(s))

	def get_database_size(self):
		''''Returns database size in MB'''
		db_size = self.sql("SELECT (pg_database_size(%s) / 1024 / 1024) as database_size",
			self.db_name, as_dict=True)
		return db_size[0].get('database_size')

	# pylint: disable=W0221
	def sql(self, *args, **kwargs):
		if len(args):
			# since tuple is immutable
			args = list(args)
			args[0] = modify_query(args[0])
			args = tuple(args)
		elif kwargs.get('query'):
			kwargs['query'] = modify_query(kwargs.get('query'))

		return super(PostgresDatabase, self).sql(*args, **kwargs)

	def get_tables(self):
		return [d[0] for d in self.sql("""select table_name
			from information_schema.tables
			where table_catalog='{0}'
				and table_type = 'BASE TABLE'
				and table_schema='public'""".format(frappe.conf.db_name))]

	def format_date(self, date):
		if not date:
			return '0001-01-01'

		if not isinstance(date, frappe.string_types):
			date = date.strftime('%Y-%m-%d')

		return date

	# column type
	@staticmethod
	def is_type_number(code):
		return code == psycopg2.NUMBER

	@staticmethod
	def is_type_datetime(code):
		return code == psycopg2.DATETIME

	# exception type
	@staticmethod
	def is_deadlocked(e):
		return e.pgcode == '40P01'

	@staticmethod
	def is_timedout(e):
		# http://initd.org/psycopg/docs/extensions.html?highlight=datatype#psycopg2.extensions.QueryCanceledError
		return isinstance(e, psycopg2.extensions.QueryCanceledError)

	@staticmethod
	def is_table_missing(e):
		return e.pgcode == '42P01'

	@staticmethod
	def is_missing_column(e):
		return e.pgcode == '42703'

	@staticmethod
	def is_access_denied(e):
		return e.pgcode == '42501'

	@staticmethod
	def cant_drop_field_or_key(e):
		return e.pgcode.startswith('23')

	@staticmethod
	def is_duplicate_entry(e):
		return e.pgcode == '23505'

	@staticmethod
	def is_primary_key_violation(e):
		return e.pgcode == '23505' and '_pkey' in cstr(e.args[0])

	@staticmethod
	def is_unique_key_violation(e):
		return e.pgcode == '23505' and '_key' in cstr(e.args[0])

	@staticmethod
	def is_duplicate_fieldname(e):
		return e.pgcode == '42701'

	@staticmethod
	def is_data_too_long(e):
		return e.pgcode == '22001'

	def create_auth_table(self):
		self.sql_ddl("""create table if not exists "__Auth" (
				"doctype" VARCHAR(140) NOT NULL,
				"name" VARCHAR(255) NOT NULL,
				"fieldname" VARCHAR(140) NOT NULL,
				"password" VARCHAR(255) NOT NULL,
				"encrypted" INT NOT NULL DEFAULT 0,
				PRIMARY KEY ("doctype", "name", "fieldname")
			)""")

	def create_global_search_table(self):
		if not '__global_search' in self.get_tables():
			self.sql('''create table "__global_search"(
				doctype varchar(100),
				name varchar({0}),
				title varchar({0}),
				content text,
				route varchar({0}),
				published int not null default 0,
				unique (doctype, name))'''.format(self.VARCHAR_LEN))

	def create_user_settings_table(self):
		self.sql_ddl("""create table if not exists "__UserSettings" (
			"user" VARCHAR(180) NOT NULL,
			"doctype" VARCHAR(180) NOT NULL,
			"data" TEXT,
			UNIQUE ("user", "doctype")
			)""")

	def create_help_table(self):
		self.sql('''CREATE TABLE "help"(
				"path" varchar(255),
				"content" text,
				"title" text,
				"intro" text,
				"full_path" text)''')
		self.sql('''CREATE INDEX IF NOT EXISTS "help_index" ON "help" ("path")''')

	def updatedb(self, doctype, meta=None):
		"""
		Syncs a `DocType` to the table
		* creates if required
		* updates columns
		* updates indices
		"""
		res = self.sql("select issingle from `tabDocType` where name='{}'".format(doctype))
		if not res:
			raise Exception('Wrong doctype {0} in updatedb'.format(doctype))

		if not res[0][0]:
			db_table = PostgresTable(doctype, meta)
			db_table.validate()

			self.commit()
			db_table.sync()
			self.begin()

	@staticmethod
	def get_on_duplicate_update(key='name'):
		if isinstance(key, list):
			key = '", "'.join(key)
		return 'ON CONFLICT ("{key}") DO UPDATE SET '.format(
			key=key
		)

	def check_transaction_status(self, query):
		pass

	def has_index(self, table_name, index_name):
		return self.sql("""SELECT 1 FROM pg_indexes WHERE tablename='{table_name}'
			and indexname='{index_name}' limit 1""".format(table_name=table_name, index_name=index_name))

	def add_index(self, doctype, fields, index_name=None):
		"""Creates an index with given fields if not already created.
		Index name will be `fieldname1_fieldname2_index`"""
		index_name = index_name or self.get_index_name(fields)
		table_name = 'tab' + doctype

		self.commit()
		self.sql("""CREATE INDEX IF NOT EXISTS "{}" ON `{}`("{}")""".format(index_name, table_name, '", "'.join(fields)))

	def add_unique(self, doctype, fields, constraint_name=None):
		if isinstance(fields, string_types):
			fields = [fields]
		if not constraint_name:
			constraint_name = "unique_" + "_".join(fields)

		if not self.sql("""
			SELECT CONSTRAINT_NAME
			FROM information_schema.TABLE_CONSTRAINTS
			WHERE table_name=%s
			AND constraint_type='UNIQUE'
			AND CONSTRAINT_NAME=%s""",
			('tab' + doctype, constraint_name)):
				self.commit()
				self.sql("""ALTER TABLE `tab%s`
					ADD CONSTRAINT %s UNIQUE (%s)""" % (doctype, constraint_name, ", ".join(fields)))

	def get_table_columns_description(self, table_name):
		"""Returns list of column and its description"""
		# pylint: disable=W1401
		return self.sql('''
			SELECT a.column_name AS name,
			CASE a.data_type
				WHEN 'character varying' THEN CONCAT('varchar(', a.character_maximum_length ,')')
				WHEN 'timestamp without TIME zone' THEN 'timestamp'
				ELSE a.data_type
			END AS type,
			COUNT(b.indexdef) AS Index,
			COALESCE(a.column_default, NULL) AS default,
			BOOL_OR(b.unique) AS unique
			FROM information_schema.columns a
			LEFT JOIN
				(SELECT indexdef, tablename, indexdef LIKE '%UNIQUE INDEX%' AS unique
					FROM pg_indexes
					WHERE tablename='{table_name}') b
					ON SUBSTRING(b.indexdef, '\(.*\)') LIKE CONCAT('%', a.column_name, '%')
			WHERE a.table_name = '{table_name}'
			GROUP BY a.column_name, a.data_type, a.column_default, a.character_maximum_length;'''
			.format(table_name=table_name), as_dict=1)

	def get_database_list(self, target):
		return [d[0] for d in self.sql("SELECT datname FROM pg_database;")]

def modify_query(query):
	""""Modifies query according to the requirements of postgres"""
	# replace ` with " for definitions
	query = query.replace('`', '"')
	query = replace_locate_with_strpos(query)
	# select from requires ""
	if re.search('from tab', query, flags=re.IGNORECASE):
		query = re.sub('from tab([a-zA-Z]*)', r'from "tab\1"', query, flags=re.IGNORECASE)

	return query

def replace_locate_with_strpos(query):
	# strpos is the locate equivalent in postgres
	if re.search(r'locate\(', query, flags=re.IGNORECASE):
		query = re.sub(r'locate\(([^,]+),([^)]+)\)', r'strpos(\2, \1)', query, flags=re.IGNORECASE)
	return query
